/*
 * File:   LCD.c
 * Author: Cicchinelli
 *
 * Created on October 27, 2018, 4:28 PM
 
 IMPORTANT:
 The functions in the file do not have any hardware dependencies.
 All of the hardware dependent functions are in LCD_IO.C
 
 */

#include "Project.h"
#include "LCD_IO.h"

// local variables also used in other modules
uint8_t	LCD_LineNbr, LCD_ColNbr;

// Send a character to the LCD
// will not print characters beyond NBR_OF_COLS
// special cases:
//		\n = goto to next line column 1
//		\r = goto column 1 of current line
//		\1..\4 = goto line x, column 1
// increments LCD_ColNbr if valid
void LCD_Char ( char x )
{
	if ( x == '\n' ) LCD_Pos ( LCD_LineNbr+1, 1 );
	else if ( x == '\r' ) LCD_Pos ( 0, 1 );
	else if ( x <= 4 ) LCD_Pos (x,1);
	else if ( LCD_ColNbr <= NBR_OF_COLS ) LCD_Data ( x ), LCD_ColNbr++;
}

void LCD_Clear (void)
{
	LCD_Cmd ( 0x01 );					// send the command
	Delay (5);							// allow LCD to respond
	LCD_LineNbr = 1;					// update line nbr
	LCD_ColNbr = 1;						//		and column nbr
} // LCD_Clear

// clear a line
// enter with: Line nbr to be cleared
//		0 = current line
void LCD_ClearLine ( uint8_t LineNbr )
{	int i;
	LCD_Pos (LineNbr,1);
	for ( i=1; i<=NBR_OF_COLS; i++) LCD_Char (' ');	
	LCD_Pos (LineNbr,1);
}

// set cursor feature
void LCD_Cursor	( uint8_t Cursor)
{	Cursor &= (CURSOR_UNDERLINE | CURSOR_BLOCK);
	LCD_Cmd ( 0x0C | Cursor );
} // LCD_Cursor

static const uint8_t LCD_LineOffset[] = {0x80+0x0, 0x80+0x40, 0x80+0x14, 0x80+0x54};

// set the LCD cursor position
// Line and Column are unsigned and 1 relative
void LCD_Pos (uint8_t line, uint8_t column )
{	uint8_t Position;

	if ( line==0 ) line = LCD_LineNbr;	// stay on same line
	if ( line>4 ) line=4;				// if invalid line nbr
	if ( column == 0 || column > NBR_OF_COLS ) column = 1;
	if ( column>NBR_OF_COLS ) column = NBR_OF_COLS;	// if invalid column nbr
	Position = LCD_LineOffset[line-1];	// create cursor position command based on line nbr
	Position += (column-1);				// insert column nbr
	LCD_Cmd ( Position );				// send to LCD
	LCD_LineNbr = line;					// save line nbr
	LCD_ColNbr = column;				//		and column nbr
} // LCD_Pos


uint8_t LCD_Line(void)		{ return LCD_LineNbr; }
uint8_t LCD_Column(void)	{ return LCD_ColNbr; }


// print characters to the LCD
// entry argument: address of text to print, text must be null terminated
void LCD_Print (char* text)
{	uint8_t i, x;
	x = strlen(text);					// get nbr of characters
	for ( i=0; i<x; i++)				// ignore null byte
		LCD_Char ( text[i] );	
} // LCD_Print

void LCD_PrintN	(char* x, uint8_t nbr)
{
	while (nbr)
	{	LCD_Char ( *x++ );
		nbr--;
	}
}
