/*
 LCD_IO.C
 This file contains all the functions that communicate directly with the
 interface to the LCD.  There are only two functions in LCD_IO that should be
 called from LCD.C: LCD_Cmd and LCD_Data

 LCD_Initialize must be called only once and only from the main program.

 This library uses commands compatible with the PCF8574T which implements the 
 equivalent of a 4 bit interface to an HD44780 via I2C.
	P0 = RS: 0=command, 1 = data
	P1 = R/W: 0 = write - this library does not implement any read operations
	P2 = EN
	P3 = LED backlight enable
	P4..P7 = D4..D7
 
 This driver requires a Delay function with 1ms resolution.
*/

#include "LCD_IO.h"

// local functions
int  I2C_Send	(uint8_t Count, uint8_t *Data);
void LCD_Nibble (uint8_t);
void LCD_Byte	(uint8_t RS_value, char x);

#define I2C_ADDR_LCD	0x27
#define NEW_LINE		0x0A

#define RS		1
#define RW		2
#define EN		4
#define BACKLIGHT 8
#define RS_CMD	0
#define RS_DATA	1


// local variables
uint8_t Bytes2Send[6];

/* Initialize the LCD
	Send all the commands required to set the LCD to:
		set the display to 4 bit mode
		set the cursor type
		Display on, blinking underline cursor
		increment cursor after write
 IMPORTANT: when using LCD_Nibble you must always send an even number
	of bytes since this library uses a 4 bit interface
*/
void LCD_Initialize()
{
// note: The 2nd nibble below become bits 0..3 if the LCD is in 4 bit data mode.
//	In this case the bits have no meaning to the LCD
//	See page 46 in the Hitachi manual
	Delay (20);
	LCD_Nibble ( 0x30 );				// 8 bit data
	Delay (10);
	LCD_Nibble ( 0x30 );				// 8 bit data - low byte if in 4 bit mode
	Delay (1);
	LCD_Nibble ( 0x30 );				// 8 bit data
	Delay (1);
	LCD_Nibble ( 0x20 );				// 4 bit data
	Delay(10);

	LCD_Byte ( RS_CMD, 0x28);			// 4 bit data and 2 lines
	LCD_Byte ( RS_CMD, 0x10);			// set cursor
	LCD_Byte ( RS_CMD, 0x0F);			// Display on, blinking underline cursor
	LCD_Byte ( RS_CMD, 0x06);			// increment cursor after write
	Delay (1);
	LCD_Clear();						// clear the display
	Delay (100);						// allow time for LCD to respond
} // LCD_Initialize

// send a command byte to the LCD
void LCD_Cmd (uint8_t cmd)
{
	LCD_Byte (RS_CMD,cmd);
}

// send a data byte to the LCD
void LCD_Data (uint8_t data)
{
	LCD_Byte (RS_DATA,data);
}

// send a byte to the LCD
// enter with:
//		RS_value = RS_DATA or RS_CMD
//		byte value to send
void LCD_Byte (uint8_t RS_value, char x)
{	uint8_t y;

	y =  (x & 0xF0) | RS_value;		// upper nibble
	y |= BACKLIGHT;
//	Bytes2Send[0] = y;
	Bytes2Send[1] = y | EN;
	Bytes2Send[2] = y;
	y = ( (x<<4) & 0xF0 ) | RS_value; // lower nibble
	y |= BACKLIGHT;
//	Bytes2Send[3] = y;
	Bytes2Send[4] = y | EN;
	Bytes2Send[5] = y;
	I2C_Send ( 6, Bytes2Send );
} // LCD_Char


// send a command nibble to the LCD
// enter with the 4 bits in bits 4..7
void LCD_Nibble (uint8_t x)
{	uint8_t y;
	y =  (x & 0xF0);
	y |= BACKLIGHT;
	Bytes2Send[0] = y;
	Bytes2Send[1] = y | EN;
	Bytes2Send[2] = y;
	I2C_Send ( 3, Bytes2Send );
} // LCD_Nibble


// send a series of bytes to the LCD
// enter with:
//		Count = number of bytes
//		Data = address of bytes to send
int I2C_Send (uint8_t Count, uint8_t *Data)
{	uint8_t i;
	uint16_t j;
	char mystring[NBR_OF_COLS];
	
	I2C1_MESSAGE_STATUS I2Cstatus;
	for (i=0;i<Count; i++) mystring[i] = Data[i];
	j = I2C_ADDR_LCD;
	
	I2C1_MasterWrite ( (uint8_t*)mystring, Count, j, &I2Cstatus);
    while ( I2C1_MasterQueueIsEmpty() == false );
	Delay(10);							// time for message to complete
	i = I2Cstatus;	// for breakpoint
	return I2Cstatus;
} // I2C_Send
