/* 
 * File:   Menu.h
 * Author: Larry_Cicchinelli
 *
 * Created on November 9, 2018, 3:27 PM
 
 IMPORTANT: neither this file nor menu.c should need to be modified by the user

 The hardware related definitions are in Menu_IO.h
 There are customization definitions in Menu_IO.h that allow you to select
	which functions you want to have available.
 The user should comment out the ones that are not going to be used
 
 user accessible variables
	these variables should be declared as extern in your file

extern uint8_t	Timeout;				// updated by users interrupt
extern char		LCDstring[21];			// general string to print to the LCD
extern char		Switch1_Name[10];
extern char		Switch2_Name[10];
extern char		Switch3_Name[10];

*******************************************************************************/
/*
Example showing usage of the Menu array

Define the menu entries
 Each entry consists of four elements:
	Menu level
	Name of function to execute if this entry is selected
	Menu entry title, 19 characters max
	Integer parameter (see below)
Notes:
	the first and last entries must be as shown below to indicate the beginning
		and end of the list: {0,0,"", n} = level 0, no function, blank name
	the top level entries are level 1
	if a level has one or more sublevels it is a "parent" and must not have a
	function.  The function address is replaced with a 0
 
Menu_t Menu[] =
{
	{0,	0, "", 1},	// top of menu list
 	{1,	0,					"Title 1", 1},
	{	2, Function_1.1,		"Sub Title 1.1", 1},
	{	2, Function_1.2,		"Sub Title 1.2", 1},
	{	2,	0,					"Sub Title 1.3", 1},
	{		3, Function_1.3.1,		"Sub Title 1.3.1", 1},
	{		3, Function_1.3.2,		"Sub Title 1.3.2", 1},
	{		3, Function_1.3.3,		"Sub Title 1.3.3", 1},
	{		3, Function_1.3.4,		"Sub Title 1.3.4", 1},
	{		3, Function_1.3.5,		"Sub Title 1.3.5", 1},
	{		3, Function_a,			"Sub Title aa", 136},
	{1,	0,					"Title 2", 1},
	{	2, Function_2.1,		"Sub Title 2.1", 1},
	{	2, Function_2.2,		"Sub Title 2.2", 1},
	{	2, Function_a,			"Sub Title ab", 23},
	{	2, Function_2.4,		"Sub Title 2.4", 1},
	{1,	0,					"Title 3", 1},
	{	2, Function_3.1,		"Sub Title 3.1", 1},
	{	2, Function_3.2,		"Sub Title 3.2", 1},
	{	2, Function_3.3,		"Sub Title 3.3", 1},
	{	2, Function_3.4,		"Sub Title 3.4", 1},
	{0, 0, "", 1}	// bottom of menu list
};
Note that a function can appear more than once in the list, even within different
	levels.  See Function_a above.

The last parameter may be used to indicate whatever the user wants to the
	executed function.  Some possibilities:
		Differentiate between the function being called by the menu system where 
			the value will be passed as the parameter vs being called
			from another function where the calling function sets the parameter value
		Indicate to the function the level of the menu from which it is being called

Required switches: see end of file for switch definitions
	Quadrature Encoder - the clock must generate an interrupt: see ext_int.c
		and must use variable QE_Int in the ISR as follows:
			if (!QE_Int) QE_Int = 1;	// show interrupt has occurred
 	SW_1		- Push Button switch on QE: select this menu entry or character
	SW_2		- accept text entry
	SW_3		- backup to previous menu level or delete previous character when entering text

Required LCD functions:
	LCD_Char	- print a single character
	LCD_Clear	- clear the display
	LCD_Cursor	- set cursor type to CURSOR_UNDERLINE, CURSOR_BLOCK or CURSOR_OFF
	LCD_Pos		- set LCD line and column - one relative
	LCD_Print	- display text
	LCD_PrintN	- display a specific number of characters
 
See Menu_IO.h for:
	macros that enable specific user-available functions
	hardware specific definitions for the switches
*/


#ifndef MENU_H
#define	MENU_H

#include "Menu_IO.h"

#define MAX_LEVEL_ENTRIES 20 // maximumm nbr of entries for any one menu level

// some useful definitions
#define	BIT_0	(1<<0)
#define	BIT_1	(1<<1)
#define	BIT_2	(1<<2)
#define	BIT_3	(1<<3)
#define	BIT_4	(1<<4)
#define	BIT_5	(1<<5)
#define	BIT_6	(1<<6)
#define	BIT_7	(1<<7)
#define	BIT_8	(1<<8)
#define	BIT_9	(1<<9)
#define	BIT_10	(1<<10)
#define	BIT_11	(1<<11)
#define	BIT_12	(1<<12)
#define	BIT_13	(1<<13)
#define	BIT_14	(1<<14)
#define	BIT_15	(1<<15)

// structure and type definition for menu entries
typedef struct
{	uint8_t	Level:3;					// limit to 7 nesting levels
	int8_t	(*Funct)(int);
	char	Name[19];
	int		Param;
} Menu_t;

/*******************************************************************************
					user accessible functions
*******************************************************************************/

// navigate through the menu
// the menu must be of type: Menu_t MyMenu[] (see typedef above))
//		see Menu.c for an example
// entry arguments: address of the Menu structure
// The selected menu function will be executed by the function Execute in main.c
// The cursor will be set to "BLOCK" and will be displayed in column 1
// The cursor will be set to "DEFAULT" and the LCD screen is cleared before calling
//		the function
void DoMenu ( Menu_t * pointer );


// get text entered by the user
// entry arguments:
//		address to put the text: should be char [21] (or less)
//		address of valid character list, null terminated
//		maximum number of characters allowed
// return value:
//		last switch used
// The cursor will be set to "UNDERLINE" when function is called
// The cursor will be set to "DEFAULT" when done
// rotating Sw1 causes the next/previous valid character to display
// pressing Sw1 on a character enters the character as part of the text,
//		advance to the next character position and displays the first
//		character of the list
// pressing Sw2 will terminate the entry
// pressing Sw3 will back up one character and delete it
//		if on the 1st character position, function will exit
// Note: the caller should display an explanatory message
//		Example:
//			LCD_Print ("Enter File Name:\n");
//			EnterText ( &FileName, &CharList, 5 );
uint8_t EnterText (char *Text, const char* ValidChars, uint8_t MaxChars);


// get a number
// the number will be displayed (and modified) at the cursor position as set by
//		the caller
// entry arguments:
//		the address of the number which should also contain the initial value
//			must be a long int
//			the value will be modified in-place
//		max nbr of digits allowed
//			caller needs to allow for '-' sign if negative values allowed
//		any text you want to print after the value, or 0 if no text
//		address of a callback function, or 0 if no callback
//			the callback will be called after each switch activation
//			the callback MUST have a single argument of long int which is the
//				updated value of the initial number
//			the cursor position will be restored upon returning
//	return value
//		1 = success
//		0 = SW3 exit
// The cursor will be set to "UNDERLINE" when function is called
// The cursor will be set to "DEFAULT" when done
// Operation:
//		the value is changed in-place and is displayed
//		pressing QE will accept the number and return it to the calling function
//		the callback will be executed upon entry to this function
//		rotating QE will increase(CW)/decrease(CCW) the number by a power of 10
//			depending on the digit at the cursor position
//			the callback will be executed
//		pressing SW_2 will move the cursor to the next more sig digit
//		pressing SW_3 will move the cursor to the next lesser sig digit
//			if already on LSdigit then exit
//		cursor left at 1 position past digits or suffix (if there is one))
int8_t GetNumber ( long int*, uint8_t, char*, void (*)() );


// select n of m options from a single line of text
// entry arguments:
//		initial option value
//		array of column nbrs of each option, terminated with 0
//			maximum nbr of items = 16
// return value:
//		updated value
// IMPORTANT: bit 0 is for the left-most option
// LCD screen must be set as desired
// The cursor will be set to "UNDERLINE" when function is called
// The cursor will be set to "DEFAULT" when done
uint16_t Select_N_on_Line ( uint16_t InitialValue, const uint8_t Columns[] );


// Enable the operator to select from among the displayed options
// The options must be on a single line
// entry argument:
//		array of column nbrs of each option, terminated with 0
//			maximum nbr of items = 8
// return value:
//		normal: the index within the array of the selection
//		SW3 exit -1;
// the text must already be displayed and the cursor must be on the text line
// navigate to the desired item and press Sw1 to select and exit
// The cursor will be set to "UNDERLINE" when function is called
// The cursor will be set to "DEFAULT" when done
int8_t Select_1_on_Line ( const uint8_t array[] );


// Enable the operator to select an entry from a list
// entry arguments:
//		a char* array containing the options as follows:
//			char *TestList[] = {
//				"entry 1",
//				"entry 2",
//				...
//				"entry n"
//			};
//			note - the text for the options do NOT have to be the same length
//			since the array passed to the function is the address of each entry
//		the number of options
//		address of title to be displayed on line 1, 0 = no title
//		IMPORTANT: all groups must be the same nbr of lines
// return value:
//		the index within the array of the selected option
//		normal: the index within the array of the selected option
//		SW3 exit: -1
// The cursor will be set to "DEFAULT" when done
// example:
//				i = sizeof(TestList)/sizeof(TestList[0]);
//				i = Select_1_of_List ( TestList, i, "Title" );
// navigate to the desired item and press Sw1 to select and exit
int8_t Select_1_of_List ( char *[], uint8_t, char* );


// get a Yes or No from the operator
// return value: 1 = yes, 0 = no
// press Sw1 to select and exit
// requires at least 6 columns remaining on the line
// must define Select_1_on_Line in Menu_IO.h
uint8_t Yes_No (void);


// convert a binary value to ASCII
// entry arguments:
//		the numeric value (long integer) to convert,
//		the number of digits
//			positive value = change leading 0s to spaces
//			negative value = keep leading 0s
//			IMPORTANT: must include space for '-' if value is negative
//				if space is not allocated, there will be no '-' sign
//		address to put the converted value
//		any prefix text or 0 if none
//		any suffix text or 0 if none
// return value:
//		index into argument 3 of trailing null byte
// this function appends a null byte to the string
// note: this function was implemented to be used in place of sprintf since the
// sprintf function added about 6KB to my code.
uint8_t Int_to_ASCII ( long int, int8_t, char*, char*, char* );


// Convert a floating point value to ASCII
// entry arguments:
//		address to put the ASCII characters, must include room for decimal point
//			and '-' if number is negative
//		total size of the converted value including '-' and decimal point
//			NOT including terminating NULL byte
//		nbr of digits to right of decimal point
//		(double)value to be converted
// exit with:
//		converted value in specified address terminated with NULL
// return value:
//		0
uint8_t FP_to_ASCII ( char*, uint8_t, uint8_t, double );


// Wait for the specified PB switch(es) - with debounce: press and release
// entry argument: the switch(es) to use
//		must be any combination of the following: SW_1 (or SW_QE), SW_2 or SW_3
// return value: value of the switch(es)
//		the return value should be used when more than 1 switch is specified
// Waits for switch release before returning to caller
uint8_t WaitPB (uint8_t);


// Wait for any switch - including the quadrature encoder
// return value: the state of the QE in bits 0,1 and switches in bits 2..4
// Waits for PB release before returning to caller
uint8_t WaitSwitches (void);


// read the Push Button switches - with debounce
// return value: value of the switch(es)
// IMPORTANT: this function does NOT wait - it reads the switches and returns
uint8_t ReadPBs (void);

void Delay		(uint16_t);				// delay a number of milliseconds

//  The hardware related definitions are in Menu_IO.h

// The following are based on how the QE and switches are used in the program.
// They should NOT be changed.
#define QE_CLK			BIT_0			// bit values
#define QE_DIR			BIT_1
#define	SW_1			BIT_2			//	 after
#define SW_2			BIT_3
#define SW_3			BIT_4			//		processing

#define SW_QE			SW_1

#define	CW				(QE_CLK|QE_DIR) // Clk = 1, Dir = 1
#define	CCW				QE_CLK			// Clk = 1, Dir = 0

#endif	// MENU_H
