/*
	main.c file initially generated from MPLAB Code Configurator

Processor: PIC 24FV32KA302

This program creates a time of day clock with up to nine alarms/timers.

ToDo:	check into removing ALARM_ON - use ALARM_EN instead
		Cycle feature: test for within cycle period when cycle time is entered
		Implement power fail detection in main loop

To be tested
		Multiple Power off-on cycles
		Cycle feature: make sure it works when spanning midnight

The number of alarms/timers may be increased but there are places in the code
where it assumes that the alarm/timer number is a single digit.  These would
have to be modified.  Search for "+'0'" if you want to find them.

Note that the sprintf function could have been used in the application in several
places; however, I found that it increased the size of my file by about 6KB.
That is why I added the Nbr_to_ASCII function to menu.c

IMPORTANT: The time values are stored as minutes past midnight.
		The largest value is 24*60 - 1 = 1439
IMPORTANT: Timer 2 and 3 are cascaded to form a 32 bit timer/counter
		All references to the timer will use timer 2
 The time is updated by an ISR in tmr2.c and is based on the CPU clock - an 8MHz
 external oscillator.  The timer is set to divide by the CPU clock frequency in
 order to generate one interrupt per second.
 Notes for tmr2.c
	must have the following
		#include "oc3.h"
		extern uint8_t	Seconds;			// seconds into current minute
		extern uint8_t	NewMinute;
		extern uint8_t	Flags;				// some communication to/from Main
		extern int		Timeout;			// menu.c
		#define SPKR_ACTIVE BIT_0			// from main.c
	The ISR must have the following code in the user area
		if ( Timeout >= 1 ) Timeout--;
		Seconds++;							// update seconds into current minute
		if ( Seconds >= 60 )				// if new minute
		{	Seconds = 0;					// reset seconds counter
			NewMinute++;					// tell new minute
		}
		if (Flags & SPKR_ACTIVE)			// if the speaker is active
		{	if ( Seconds & 1 ) OC3_Start();	// turn on speaker PWM
			else OC3_Stop();				// turn off speaker PWM
		}

Operation Options
	Time of day can be in either 12 Hour (AM/PM) or 24 Hour mode
	Time of day clock,  with an option to display seconds,
		& date with up to 7 alarm/timer settings
	Alarms/Timers 1-7: one-time or repeated based on selected days of week
 The Alarms/Timers can utilize up to four outputs:
	Speaker with programmable tone frequency (resolution = 8Hz) up to 2KHz
	One, built-in DPDT relay
	Four open drain outputs which can be used to drive external devices such as
		relays and/or triacs.  One output is shared with the relay.
*/

#ifdef DEBUG			// Project.h
#warning DEBUG is defined
#endif

/*
  Section: Included Files
*/
#include "Project.h"

/******************************************************************************/
/********************* EEPROM Definitions and Memory Usage ********************/
/******************************************************************************/

#define wEE_SIZE			256			// these four values are from the
#define EE_BASE				0x7FFE00	// 24FV32KA301 manual
#define EE_PAGE				0x7F
#define EE_OFFSET			0xFE00

// This value is determined by the way the EEPROM must be accessed
#define WORDS_PER_BLOCK		4			// 4 words per block

/******************************************************************************/
/********************************* System Data ********************************/
/******************************************************************************/

/* This structure is read from the System blocks of the EEPROM.  The total size
 of the structure must be a multiple of four words (eight bytes) due to the way
 data is written to the EEPROM.  Since the compiler does not have a directive
 that tells the size of the structure it has to be determined manually by
 totaling the sizes of all elements.
 Current size is 21 bytes ==> 12 words = 24 bytes = 3 blocks
*/
struct
{ // IMPORTANT: Tmr2_Divlo MUST be first: see Read_Alarm and Write_Alarm
	// put words on even address
	uint16_t	Tmr2_Divlo;		// Timer divisor low order word
	uint16_t	Tmr2_Divhi;		// Timer divisor high order word
	uint16_t	Bright_tm_day;	// daytime brightness time of day
	uint16_t	Bright_tm_nite;	// dnitetime brightness time of day
	struct
	{	uint16_t day	: 5;	// day of month: 1-31
		uint16_t month	: 4;	// month of year: 1-12
		uint16_t year	: 7;	// year offset from 2019
	}Date;
	// put bytes starting here
	uint8_t Token;
	struct
	{	uint8_t Hr24	: 1;	// 1 = 24 Hour format
		uint8_t Cal		: 1;	// 1 = calibrated
		uint8_t DST_En	: 1;	// 1 = DST enabled
		uint8_t DST		: 1;	// 1 = currently in DST
		uint8_t Secs	: 1;	// 1 = display seconds
		uint8_t Format	: 1;	// 1 = D/M/Y, 0 = M/D/Y	DATE FORMAT
	}Flag;
	uint8_t		Snooze;			// snooze time in minutes
	uint8_t		Bright_Day;		// daytime brightness level
	uint8_t		Bright_Nite;	// nighttime brightness level
	uint8_t		EZ_Days;		// EZ alarm: days of week for 
	uint8_t		EZ_Tone;		// EZ alarm: tone frequency
	uint8_t		EZ_Stop;		// EZ alarm: stop offset in minutes
	uint8_t		EZ_Repeat;		// EZ alarm: repeat weekly flag
	uint8_t		TMR_Tone;		// EZ timer: tone frequency
	uint8_t		TMR_Stop;		// EZ timer: stop offset in minutes
} Sys;

#define TOKEN_VALUE			0x0A		// indcate that EEPROM has been initialized

#define NBR_SYS_BLOCKS		3
#define wSYS_SIZE			(WORDS_PER_BLOCK*NBR_SYS_BLOCKS)
#define bSYS_SIZE			(2*wSYS_SIZE)

/******************************************************************************/
/****************************** Alarm/Timer Data ******************************/
/******************************************************************************/

// Alarms is the alarm/timer data read from/written to the EEPROM

#define NBR_OF_ALARMS		9			// alarms and timers

/* This structure is read from the Alarm blocks of the EEPROM.  The total size
 of the structure must be a multiple of four words (eight bytes) due to the way
 data is written to the EEPROM.  Since the compiler does not have a directive
 that tells the size of the structure it has to be determined manually by
 totaling the sizes of all elements.
 Current size is 10 bytes ==> 8 words = 16 bytes = 2 blocks
*/
struct
{ // IMPORTANT: Start MUST be first: see Read_Alarm and Write_Alarm
	// put words on even address

	uint16_t Start;				// start time of the Alarm/timer
	uint16_t Stop;				// stop time
	struct
	{	uint8_t Repeat	: 1;	// 1 = Alarm/Timer repeats weekly
		uint8_t EZ_Timer : 1;	// 1 = this is an Easy Timer
		uint8_t On		: 1;	// 1 = Alarm/Timer is On
		uint8_t Cycle	: 1;	// 1 = Timer has on-off cycles
		uint8_t Enabled	: 1;	// 1 = Alarm/timer is enabled
		uint8_t Setup	: 1;	// 1 = Alarm/timer has been setup
	}Mode;

	union		// union is needed in order to be able to access the entire byte
	{	struct
		{	uint8_t Spkr	: 1;	// 1 = speaker is enabled
			uint8_t O1		: 1;	// 1 = output 1 is enabled
			uint8_t O2		: 1;	// 1 = output 2 is enabled
			uint8_t O3		: 1;	// 1 = output 4 is enabled
			uint8_t O4		: 1;	// 1 = output 4 is enabled
		}f;			// individual bits
		uint8_t b;	// entire byte
	} Outputs;
	uint8_t Days;				// days of the week that Alarm/Timer is active
	uint8_t Tone;				// Alarm tone frequency/8
	uint8_t Cycle_On;			// Timer cycle on time
	uint8_t Cycle_Off;			// Timer cycle off time
}Alarms[NBR_OF_ALARMS];

#define BLOCKS_PER_ALARM	2

#if BLOCKS_PER_ALARM*8 < sizeof(Alarms[0] )
#error Too much data in Alarms array
#endif

#define START		Alarms[AlrmTmrNbr-1].Start
#define STOP		Alarms[AlrmTmrNbr-1].Stop
#define MODE		Alarms[AlrmTmrNbr-1].Mode
#define OUTPUTS		Alarms[AlrmTmrNbr-1].Outputs
#define DAYS		Alarms[AlrmTmrNbr-1].Days
#define TONE		Alarms[AlrmTmrNbr-1].Tone
#define CYCLE_ON	Alarms[AlrmTmrNbr-1].Cycle_On
#define CYCLE_OFF	Alarms[AlrmTmrNbr-1].Cycle_Off
#define SETUP		Alarms[AlrmTmrNbr-1].Mode.Setup

#define FIRST_ALARM_BLOCK	NBR_SYS_BLOCKS
#define NBR_ALARM_BLOCKS	(NBR_OF_ALARMS*BLOCKS_PER_ALARM)

#define wALARM_SIZE			(WORDS_PER_BLOCK*BLOCKS_PER_ALARM)
#define bALARM_SIZE			(2*wALARM_SIZE)

#define EE_BLOCKS_USED	( NBR_SYS_BLOCKS + (NBR_OF_ALARMS * BLOCKS_PER_ALARM) )
#if (EE_BLOCKS_USED*WORDS_PER_BLOCK) > wEE_SIZE
#error Too much EEPROM allocated
#endif

#define NBR_OUTPUTS 4

/******************************************************************************/
/******************* End EEPROM Definitions and Memory Usage ******************/
/******************************************************************************/

#define LCD_PERIOD			4000		// period of LCD brightness PWM
#define NOON				720			// 12*60 = 720 minutes past midnight
#define MPD					1440		// minutes per day
#define MPW					10080		// Minutes per week

#if CYCLE_RESOLUTION > 6
#error CYCLE_RESOLUTION too large
#endif

// A/D channel definitions
#define AD_BATT		0
#define AD_VIN		1
#define AD_MULT		3.0					// voltage divider ratio
#define MIN_BATT	6.0					// minimum allowed battery voltage
#define MIN_VIN		6.5					// minimum allowed main voltage

// Global variables used from other modules
extern char LCDstring[21];				// menu.c - general usage

// the following three variables are also used in tmr2.c
uint8_t		Flags;							// misc communication bits
	#define FLAG_RINGING	BIT_0
	#define FLAG_PM			BIT_7		// 1=PM, 0=AM
	#define FLAG_LOW_BATT	BIT_6		// battery low
	#define FLAG_PWR_FAIL	BIT_5		// main power down

uint8_t		Seconds;					// seconds into current minute
uint8_t		NewMinute;

/******************************************************************************/
// Global variables used in this module
uint8_t		AlrmTmrNbr;					// used by the EE_ macros
int			CurrentMins;				// minutes past midnight
uint8_t		DayOfWeek;					// day of week, 0=Sunday
uint8_t		DayOfMonth;
uint8_t		Month;						// 1 = Jan
uint16_t	Year;
long int	CalMins;					// nbr of minutes since the last Time set
char		DispLine[21];
long int	SnoozeTime;
uint8_t		TimeLength;					// nbr of characters to display the time
long int	CycleMinutes;
uint16_t	CycleTimes[NBR_OF_ALARMS][2]; // temp store of begin and end times
long int	ToneFrequency;				// speaker frequency

// power failure data
#define MAX_POWER_OFF_CYCLES	10
uint8_t		PowerOffCount;
typedef struct
{	int8_t		OffDay;
	int16_t		OffTime;
	int8_t		OnDay;
	int16_t		OnTime;
} Power_t;
Power_t Power[MAX_POWER_CYCLES];

uint8_t	AlarmFlags;
	#define	AF_SNOOZE		BIT_7		// snooze is in progress
	#define	AF_ALARMS_EN	BIT_6		// 1 = alarms enabled this cycle

uint8_t	ActiveAlrmTmrNbr;				// alarm/timer nbr being processed 1 relative
uint8_t NextAlarmNbr;
uint8_t SnoozeAlarmNbr;

/******************************************************************************/

const uint8_t DaysPerMonth[12] = {31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31};
//								  Jan Feb Mar Apr May Jun Jul Aug Sep Oct Nov Dec
const char MonthNames[12][5] =
{ "Jan ", "Feb ", "Mar ", "Apr ", "May ", "Jun ", "Jul ", "Aug ", "Sep ", "Oct ", "Nov ", "Dec " };
const char DayNames[7][5] = { "Sun,", "Mon,", "Tue,", "Wed,", "Thu,", "Fri,", "Sat," };
const char DaysShort[] =  "Su Mo Tu We Th Fr Sa ";
const uint8_t DaysShort_Cols[] = {1,4,7,10,13,16,19,0};

// the delay functions are declared in ClkTmr24.h for convenience

// Local function declarations
void	MainDisplay		(void);
void	Display_Time	(int,uint8_t);
void	Set_LCD_Brightness (long int);
void	Set_Spkr_Freq	(long int);
int		Enter_Time		(int);
void	Display_Date	(void);
void	Handle_New_Minute (void);
void	Start_Timer		(uint8_t);
void	Check_Alarms	(void);
void	NextAlarm		(void);
void	Display_Alarm	(int8_t);
uint8_t	Display_All_Alarms (void);
void	CheckPower		(uint8_t);
double	ReadAD			(uint8_t Channel);
void	LCD_Bright_Day_Night (void);
void	ShowVoltages	(void);
void	FP2Ascii		(double, uint8_t);

// EEPROM functions
void	Read_Alarm		(int8_t);
void	Write_Alarm		(int8_t);

/******************************************************************************
							Menu
 ******************************************************************************/

// Menu execute function declarations
// Easy
int8_t Mnu_Alarm_Easy	(int);
int8_t Mnu_1_Time_Alarm	(int);
// Alarms/Timers
int8_t Mnu_Full			(int);
int8_t Mnu_Display		(int);
	// Detailed Setup
int8_t Mnu_Time			(int);
int8_t Mnu_Days			(int);
int8_t Mnu_Outputs		(int);
int8_t Mnu_Tone			(int);
int8_t Mnu_Enable		(int);
int8_t Mnu_Repeat		(int);
int8_t Mnu_Cycle		(int);
int8_t Mnu_Clear		(int);
int8_t Mnu_Save			(int);
// Current Time/Date
int8_t Mnu_ToD			(int);
int8_t Mnu_24hr			(int);
int8_t Mnu_DOW			(int);
int8_t Mnu_Date			(int);
int8_t Mnu_Cal			(int);
// EZ
int8_t Mnu_EZ_Al_Days	(int);
int8_t Mnu_EZ_Al_Tone	(int);
int8_t Mnu_EZ_Al_Stop	(int);
int8_t Mnu_EZ_Al_Rpt	(int);
int8_t Mnu_EZ_Tmr_Stop	(int);
int8_t Mnu_EZ_Tmr_Tone	(int);
int8_t Mnu_EZ_Save		(int);
// utilities
int8_t Mnu_Snooze		(int);
int8_t Mnu_LCD_Bright	(int);
int8_t Mnu_Timer_Div	(int);
int8_t Mnu_Power		(int);
int8_t Mnu_DST			(int);
int8_t Mnu_Disp_Secs	(int);
int8_t Mnu_Date_Format	(int);
int8_t Mnu_Info			(int);

const Menu_t MyMenu[] =					// Menu_t is defined in menu.h
// note: the indenting is strictly for readability
{
	{0, 0, "", 0},	// top  of menu list
	{1, Mnu_Alarm_Easy,		"Easy Alarm", 1},
	{1, Mnu_1_Time_Alarm,	"1 Time Alarm: Mins", 1},
	{1, Mnu_1_Time_Alarm,	"1 Time Alarm: ToD", 2},
	{1,	0,					"Alarms/Timers", 0},
	{	2, Mnu_Full,			"Full Setup", 1},
	{	2, 0,					"Detailed Setup", 1 },
	{		3, Mnu_Time,			"Set Times", 1},
	{		3, Mnu_Days,			"Select Days of Week", 1},
	{		3, Mnu_Outputs,			"Select Outputs", 1},
	{		3, Mnu_Tone,			"Set Tone", 1},
	{		3, Mnu_Repeat,			"Repeat Weekly", 1},
	{		3, Mnu_Enable,			"Enable/Disable", 1},
	{		3, Mnu_Cycle,			"Cyclic On-Off", 1},
	{		3, Mnu_Clear,			"Clear", 1},
	{		3, Mnu_Save,			"Save?", 1},
	{	2, Mnu_Display,			"Display All Alarms", 1},
 	{1,	0,					"Current Time/Date", 0},
	{	2, Mnu_ToD,				"Time", 1},
	{	2, Mnu_24hr,			"AM/PM or 24Hr", 1},
	{	2, Mnu_DOW,				"Day of Week", 1},
	{	2, Mnu_Date_Format,		"Date Format", 1},
	{	2, Mnu_Date,			"Date", 1},
	{	2, Mnu_Cal,				"Calibrate", 1},
	{1, 0,					"Easy Defaults", 1 },
	{	2, Mnu_EZ_Al_Days,		"EZ Alarm Days", 1 },
	{	2, Mnu_EZ_Al_Tone,		"EZ Alarm Tone", 1 },
	{	2, Mnu_EZ_Al_Stop,		"EZ Alarm Stop", 1 },
	{	2, Mnu_EZ_Al_Rpt,		"EZ Alarm Repeat", 1 },
	{	2, Mnu_EZ_Tmr_Stop,		"EZ Timer Stop", 1 },
	{	2, Mnu_EZ_Tmr_Tone,		"EZ Timer Tone", 1 },
	{	2, Mnu_EZ_Save,			"EZ Save", 1 },
	{1,	0,					"Utilities", 0},
	{	2, Mnu_Snooze,			"Snooze Time", 1},
	{	2, Mnu_LCD_Bright,		"LCD Brightness", 1},
	{	2, Mnu_Timer_Div,		"Timer Divisor", 1},
	{	2, Mnu_Power,			"Power", 1},
	{	2, Mnu_DST,				"DST", 1 },
	{	2, Mnu_Disp_Secs,		"Show Seconds", 1 },
	{1, Mnu_Info,			"Build Info", 1 },
	{0, 0, "", 0}	// end  of menu list
};

/******************************************************************************/

/*******************************************************************************
Main application
*******************************************************************************/
int main(void)
{	uint32_t Divisor;
	int LastSecond;
	uint16_t i, sw_nbr;
	uint8_t MainFlags;
#define FORCE_UPDATE	BIT_0

// initialize the device
	Seconds = LastSecond = 0;
	NewMinute = 0;
	ActiveAlrmTmrNbr = SnoozeAlarmNbr = SnoozeTime = 0;
	NextAlarmNbr = AlarmFlags = 0;
	SYSTEM_Initialize();
	Set_LCD_Brightness (50);
	LCD_Initialize();
	OC2_SecondaryValueSet ( LCD_PERIOD );
	Set_Spkr_Freq (400);
	OC3_Stop();							// stop speaker output
	OC3pin = 0;							// force speaker output low when no PWM
	PowerOffCount = 0;					// no power off events
	LCD_Cursor(CURSOR_DEFAULT);

// read/check EEPROM
	MainFlags = 0;
	Read_Alarm(0);						// read the system data
	if ( Sys.Token != TOKEN_VALUE )		// if EEPROM has NOT been initialized
	{
	// initialize system data with all default values
		memset ( &Sys, 0, bSYS_SIZE );
		LCD_Clear();
		LCD_Print ( "Initializing System\n   Default values\n" );
		Delay (1000);
		Sys.Token = TOKEN_VALUE;
		Divisor = _XTAL_FREQ/2;			// timer clock from main oscillator
		TMR2_Period32BitSet( Divisor );	// set timer 2
		Sys.Tmr2_Divhi = Divisor>>16;	// save MSW
		Sys.Tmr2_Divlo = Divisor;		// save LSW
		Sys.Flag.DST_En = 1;			// enable DST detection
		Sys.Snooze = DEF_SNOOZE_TIME;	// default snooze time = 10 minutes
		// set LCD brightness values
		Sys.Bright_tm_day = BRIGHT_TM_DAY;
		Sys.Bright_Day = BRIGHT_DAY;
		Sys.Bright_tm_nite = BRIGHT_TM_NITE;
		Sys.Bright_Nite = BRIGHT_NITE;
		// set the Easy values
		Sys.EZ_Days = EASY_DAYS;
		Sys.EZ_Tone = EASY_TONE/8;
		Sys.EZ_Stop = EASY_ALARM_STOP;
		Sys.EZ_Repeat = EASY_REPEAT;
		Sys.TMR_Tone = EASY_TIMER_TONE/8;
		Sys.TMR_Stop = EASY_TIMER_STOP;
		Mnu_24hr(0);					// get 12/24 hour flag
		Mnu_Date_Format(0);				// and date format
		Write_Alarm(0);					// save preliminary system data
		memset ( &Alarms[0], 0, NBR_OF_ALARMS*bALARM_SIZE );
										// clear the alarm/timer data
	} // if EEPROM has NOT been initialized

	else // if EEPROM HAS been initialized
	{	for (AlrmTmrNbr=1; AlrmTmrNbr<=NBR_OF_ALARMS; AlrmTmrNbr++ )
			Read_Alarm (AlrmTmrNbr);	// read each alarm/timer
	}

// get time and date
	CurrentMins = MPD/2;				// noon, initial value for Mnu_ToD
	Mnu_ToD(0);							// also starts the timer
	Mnu_DST(0);
	Mnu_DOW(0);
	Mnu_Date(0);
	if ( Sys.Flag.Hr24 ) TimeLength = 5; // 24hr value
	else TimeLength = 8;				// AM/PM value
	if ( Sys.Flag.Secs ) TimeLength += 3;
	LCD_Bright_Day_Night();
	MainFlags |= FORCE_UPDATE;
	AlarmFlags = AF_ALARMS_EN;			// default to alarms enabled

/******************************************************************************/
/******************************* main loop ************************************/
/******************************************************************************/

	while (1)
	{	LCD_Cursor ( CURSOR_OFF );
		if ( NewMinute )
		{	Handle_New_Minute();
			Check_Alarms();
			MainFlags |= FORCE_UPDATE;
		} // if new minute
		if ( MainFlags & FORCE_UPDATE ) // if update requested
		{	MainFlags &= ~FORCE_UPDATE; // clear flag for next time
 			MainDisplay();
		}
		CheckPower(2);					// test for main power
		sw_nbr = ReadPBs();
		if ( sw_nbr )					// if a switch has been pressed
		{	while ( ReadPBs() );		// wait for switch release
			if ( Flags & FLAG_RINGING ) // if alarm is "ringing"
			{ // any SW will turn off the alarm
				Flags &= ~FLAG_RINGING;	// disable ISR from turning it on/off
				OC3_Stop();				// turn off tone
				if ( sw_nbr == SW_2 )	// set up for SNOOZE
				{	SnoozeTime = Sys.Snooze; // get snooze minutes
					AlarmFlags |= AF_SNOOZE; // show snoozing
					AlrmTmrNbr = SnoozeAlarmNbr; // needed by MODE
					// SnoozeAlarmNbr was set in Check_Alarms
					MODE.On = 1;		// allow alarm to ring
				}
				else					// SW_1 or SW_3
				{	AlrmTmrNbr = SnoozeAlarmNbr; // needed for MODE
					SnoozeAlarmNbr = 0;	// used in MainDisplay
					AlarmFlags &= ~AF_SNOOZE; // show not snoozing
					if ( MODE.EZ_Timer ) // clear timer info
						memset ( &Alarms[AlrmTmrNbr-1], 0, bALARM_SIZE );
				}
				MainFlags |= FORCE_UPDATE;
			}
			else if ( sw_nbr == SW_QE )	// do NOT move this if clause
			{
			/************************************************/
				DoMenu( (Menu_t *)MyMenu );
				// the reason for casting the type is because the menu in this
				// file is declared as a const Menu_t whereas Menu.c only requires
				// Menu_t.  This allows the menu to be in RAM if desired.
			/************************************************/
				MainFlags |= FORCE_UPDATE;
			}
			else if ( (sw_nbr == SW_2) && NextAlarmNbr )
			{	AlarmFlags &= ~AF_SNOOZE; // force no snooze
				SnoozeAlarmNbr = SnoozeTime = 0; // show no alarm is snoozing
				AlarmFlags ^= AF_ALARMS_EN; // invert alarms enabled this cycle
				MainFlags |= FORCE_UPDATE;
				AlrmTmrNbr = NextAlarmNbr; // needed for MODE
				if ( MODE.EZ_Timer )	// clear timer info
					memset ( &Alarms[AlrmTmrNbr-1], 0, bALARM_SIZE );
			}
		} // if sw

		if ( Sys.Flag.Secs )
		{	i = (NBR_OF_COLS - TimeLength)/2;
			LCD_Pos (1,i+1);
			Display_Time ( CurrentMins, 1 );
		}
		else							// if NOT displaying seconds
		{	if ( LastSecond != Seconds  )
			{	LastSecond = Seconds;
//				LCD_Pos (1,1);
				LCD_Print ( "\1  \r" );
				if ( Seconds & 1 ) LCD_Pos ( 1,2 );
				LCD_Char ( '.' );
			}
		}
//		if ( ReadAD(AD_VIN) < MIN_VIN ) MainFlags |= FORCE_UPDATE;
	} // while (1) - end of main loop

/******************************************************************************/
/****************************** end of main loop ******************************/
/******************************************************************************/

	return 1;	// this should never get executed
} // main


// display Time, Date and Alarm info
// Line 1: time
// Line 2: date
// Line 3: battery state
// Line 4: alarm info
void MainDisplay (void)
{	uint8_t i;

	LCD_Clear();
	if ( PowerOffCount ) LCD_Print ("P");
	i = (NBR_OF_COLS - TimeLength)/2;
	if ( !Sys.Flag.Secs )
	{	LCD_Pos (1,i+1);
		Display_Time ( CurrentMins, 1 );// display time
	}
	LCD_Pos (1,20);
	if ( Sys.Flag.DST ) LCD_Char ( 'D' );
	else LCD_Char ( 'S' );
	LCD_Pos (2,1);
	Display_Date();						// date
	LCD_Pos (3,1);
	CheckPower(3);
	LCD_Pos (4,1);
	if ( AlarmFlags & AF_SNOOZE )
	{	LCD_Char ('A' ), LCD_Char ( SnoozeAlarmNbr +'0' );
		Nbr_to_ASCII ( (long)SnoozeTime, 2, &LCDstring[0], " Snz ", 0 );
		LCD_Print ( LCDstring );
	}
	else NextAlarm();					// and next alarm
} // MainDisplay

// find the next active alarm clock
// the LCD cursor must be set as desired
void NextAlarm(void)
{	uint8_t Mode, Day, ThisDay, i, alarm_day;
	int16_t Mins2ClosestAlarm, Mins2Alarm;

	Mins2ClosestAlarm = MPW+1;
	NextAlarmNbr = 0;

	for ( i=DayOfWeek; i<DayOfWeek+7; i++) // for each day of the week
	{	Day = i%7;						// day nbr
		for (AlrmTmrNbr=1; AlrmTmrNbr<=NBR_OF_ALARMS; AlrmTmrNbr++ ) // for each alarm
		{	Mode = MODE.On;				// !0 = alarm is on
			ThisDay = DAYS & (1<<Day);	// !0 = alarm is for this day
			Mins2Alarm = MPW+2;
			if ( OUTPUTS.f.Spkr && Mode && ThisDay )
			{	if ( Day == DayOfWeek )
				{	if ( START > CurrentMins ) Mins2Alarm = START - CurrentMins;
				}
				else Mins2Alarm = START;
				if ( Mins2Alarm < Mins2ClosestAlarm )
				{	Mins2ClosestAlarm = Mins2Alarm;
					alarm_day = Day;
					NextAlarmNbr = AlrmTmrNbr;
				}
			} // if alarm is active for this day
		} // for each alarm
		if ( NextAlarmNbr ) break;		// done if alarm found
	} // for each day of the week

	AlrmTmrNbr = NextAlarmNbr;			// get alarm nbr for macros
	if ( AlrmTmrNbr )
	{	LCD_Print ( "A" );
		LCD_Char ( AlrmTmrNbr +'0' );
		if ( (AlarmFlags & AF_ALARMS_EN) ) LCD_Print ( " On" );
		else LCD_Print ( " Off");
		i = 10;							// starting column if AM/PM display
		if ( Sys.Flag.Hr24 ) i = 13;	// if 24 hour display
		if ( START < NOON ) i += 1;
		LCD_Pos ( 0, i );
		LCD_PrintN ( (char*)&DaysShort[alarm_day*3], 3 );
		Display_Time ( START, 0 );
	}
	else
		LCD_Print ("no Alarm");
} // NextAlarm


/*******************************************************************************
							Current Time Functions
*******************************************************************************/

// select between a 24 hour display vs a 12 hour AM/PM display
// this function either sets or clears the Hr24 bit in Flags
const uint8_t Indexes_24Hr[] = {3,12,0};
int8_t Mnu_24hr (int MenuParam)
{	int8_t i, j;

	LCD_Clear();
	LCD_Print ("AM/PM or 24Hour?\r");
			//	    ^		 ^
			//      3   7    12
	LCD_Cursor(CURSOR_BLOCK);
	i = Select_1_on_Line ( Indexes_24Hr );
	j = Sys.Flag.Hr24;					// save current value
	if ( i != 0 ) TimeLength = 5, Sys.Flag.Hr24 = 1; // set to 24Hr
	else TimeLength = 8, Sys.Flag.Hr24 = 0; // set to 12Hr
	if ( j != Sys.Flag.Hr24 ) Write_Alarm(0); // save if value has changed
	if (Sys.Flag.Secs) TimeLength += 3;
	LCD_Cursor ( CURSOR_DEFAULT );
	return 1;
} // Mnu_24hr

int8_t Mnu_ToD (int MenuParam)
{
	LCD_Clear();
	LCD_Print ("Enter Time of Day:\n      ");
	TMR2_Stop();
	CurrentMins = Enter_Time ( CurrentMins );
	Start_Timer(1);
	return 1;
} // Mnu_ToD


int8_t Mnu_DOW (int MenuParam)
{
	LCD_Clear();
	LCD_Print ("Set the current\nDay of the Week:\n");
	LCD_Print ( (char*)DaysShort );
	DayOfWeek = Select_1_on_Line ( DaysShort_Cols );
	return 1;
} // Mnu_DOW


// get current date
#define COLUMN_1	6
#define COLUMN_2	(COLUMN_1+3)
#define COLUMN_3	(COLUMN_1+8)
int8_t Mnu_Date (int MenuParam)
{	int8_t CurrentCol, i;
	uint8_t SwValue, LastDOM, value1, value2;
	char DateString[11];
	uint8_t *Val_1, *Val_2;

	LCD_Clear();
	LCD_Print ("Enter Date as");
							// 123456 +3   +8
	DayOfMonth = 1;
	Month = 1;
	Year = 2019;
	CurrentCol = COLUMN_1;				// point to LS digit

	if ( !Sys.Flag.Format )				// mm/dd/yyyy
	{	Val_1 = &Month;
		Val_2 = &DayOfMonth;
		LCD_Print ("\n    mm/dd/yyyy\n");
				//    123456 +3   +8
	}
	else								// dd/mm/yyyy
	{	Val_2 = &Month;
		Val_1 = &DayOfMonth;
		LCD_Print ("\n    dd/mm/yyyy\n");
				//    123456 +3   +8
	}

	while (1)
	{	LastDOM = DaysPerMonth[Month-1]; // get last day of month
		if ( (Month==2) && (Year%4==0) ) LastDOM = 29; // if Feb. of leap year

		value1 = Month, value2 = DayOfMonth;
		if ( Sys.Flag.Format ) value1 = DayOfMonth, value2 = Month;
		i = Nbr_to_ASCII ( (long int) value1, 2, &DateString[0], 0, "/" );
		i += Nbr_to_ASCII ( (long int) value2, 2, &DateString[i], 0, "/" );
		Nbr_to_ASCII ( (long int) Year, 4, &DateString[i], 0, 0 );

		LCD_Pos (0,COLUMN_1-1);			// position of MS digit
		LCD_Print ( DateString );
		LCD_Pos (0,CurrentCol);
		SwValue = WaitSwitches();		// wait for a switch
		if ( SwValue == CW )
		{	if (CurrentCol==COLUMN_1) *Val_1 = *Val_1 + 1;
			if (CurrentCol==COLUMN_2) *Val_2 = *Val_2 + 1;
			if (CurrentCol==COLUMN_3) Year++;
		}
		else if ( SwValue == CCW )
		{	if (CurrentCol==COLUMN_1) *Val_1 = *Val_1 - 1;
			if (CurrentCol==COLUMN_2) *Val_2 = *Val_2 - 1;
			if (CurrentCol==COLUMN_3) Year--;
		}
		else if ( SwValue==SW_2 )
		{	if (CurrentCol==COLUMN_1) CurrentCol = COLUMN_3;
			else if (CurrentCol==COLUMN_2) CurrentCol = COLUMN_1;
			else CurrentCol = COLUMN_2;
		}
		else if ( SwValue == SW_QE )
		{	if (CurrentCol==COLUMN_1) CurrentCol = COLUMN_2;
			else if (CurrentCol==COLUMN_2) CurrentCol = COLUMN_3;
			else break;
		}
		if ( Month > 12 ) Month = 1;
		if ( Month < 1 ) Month = 12;
		if ( DayOfMonth > LastDOM ) DayOfMonth = 1;
		if ( DayOfMonth < 1 ) DayOfMonth = LastDOM;
		if ( Year < 2019 ) Year = 2019;
	}
	return 1;
} // Mnu_Date


// Calibrate the clock
// Both the starting and the ending times MUST be entered
// when your clock standard shows 00 seconds into the minute.
// displayed values: original Timer divisor, new Timer divisor
//		seconds in calibration interval, Error in seconds, correction value
int8_t Mnu_Cal (int MenuParam)
{	long int TmrPeriod0, TmrPeriod1;
	long int CalSeconds;
	long int ActMins, ErrSeconds, i;
	float Correction, aa, bb, cc;
	uint8_t j;

// get current timer 2 period
	TmrPeriod0 = PR3;					// get high order word
	TmrPeriod0 <<= 16;					// put into high word
	TmrPeriod0 += PR2;					// add low order word
// get new time
	LCD_Print("Act. time: ");
	ActMins = Enter_Time ( CurrentMins ); // get actual time
	ErrSeconds = (CurrentMins - ActMins) * 60 + Seconds;
	CalSeconds = (CalMins * 60) + Seconds; // seconds in calibration interval
// correction =  Timer 2 Period * (error seconds/seconds in calibration period)
	aa = TmrPeriod0;					// convert
	bb = ErrSeconds;					//	 to float
	cc = CalSeconds;					//		for calculation
	Correction = aa*(bb/cc);			// correction > 0 if clock is fast
	i = Correction;
	TmrPeriod1 = TmrPeriod0 + (long)i;	// tmr2 period must be increased
										//		if clock is fast
// display values: original Timer divisor, new Timer divisor
	j = Nbr_to_ASCII ( TmrPeriod0, 8, &LCDstring[0], "\n", " " );
	Nbr_to_ASCII ( TmrPeriod1, 8, &LCDstring[j], 0, "\n" );
	LCD_Print (LCDstring);

// display: seconds in calibration interval, Error in seconds, correction value
	j = Nbr_to_ASCII ( CalSeconds, 7, &LCDstring[0], 0, " " );
	j += Nbr_to_ASCII ( ErrSeconds, 4, &LCDstring[j], 0, " " );
	Nbr_to_ASCII ( i, 4, &LCDstring[j], 0, "\n" );

	LCD_Print (LCDstring);
//	LCD_Pos ( 4,1 );
	LCD_Print ("\4Save? ");
	if ( !Yes_No() ) return 0;

// save new period values and re-start timer
// set new timer 2 period
	PR3 = TmrPeriod1;					// set high order word
	TmrPeriod1 <<= 16;					// put into high word
	PR2 = TmrPeriod1;					// set low order word
	Sys.Tmr2_Divhi = TmrPeriod1>>16;	//	save new
	Sys.Tmr2_Divlo = TmrPeriod1;		//		divisor
	Sys.Flag.Cal = 1;					// show system had been calibrated
	Write_Alarm (0);
	LCD_Print ("Calibration Stored\n");
	Mnu_ToD(0);
	return 1;
} // Mnu_Cal


/*******************************************************************************
						Alarm/Timer Menu Functions
*******************************************************************************/

// easy alarm using mostly pre-defined values
// see Project.h for the default values
int8_t Mnu_Alarm_Easy (int MenuParam)
{
	for ( AlrmTmrNbr=1; AlrmTmrNbr<NBR_OF_ALARMS; AlrmTmrNbr++ )
		if ( !MODE.Setup ) break;			// if alarm not used
	if ( AlrmTmrNbr >= NBR_OF_ALARMS )
	{	LCD_Print ( "\nNo alarms available!");
		Delay ( 1000 );
		return 0;
	}
	LCD_Print ( "Using Alarm ");
	LCD_Char ( AlrmTmrNbr+'0' );
	LCD_Print ( "\nEnter Alarm Time:\n    ");
	START = Enter_Time (CurrentMins);
	if ( (int16_t)START < 0 ) return 0;
	if ( !(START & BIT_15) )
	{	MODE.Enabled = MODE.On = MODE.Setup = 1;
		if ( Sys.EZ_Repeat ) MODE.Repeat = 1;
		OUTPUTS.f.Spkr = 1;
		DAYS = Sys.EZ_Days;				// 0011 1110, Mon-Fri, bit 0 = Sunday
		TONE = Sys.EZ_Tone/8;
		STOP = START+Sys.EZ_Stop;
		if ( STOP > MPD ) STOP -= MPD;
		Write_Alarm ( AlrmTmrNbr );
	}
	return 1;
} // Mnu_Alarm_Easy


// EZ Timer using mostly pre-defined values
// see Project.h for the default values
// MODE will be cleared when the alarm rings or when it is stopped via SW2
int8_t Mnu_1_Time_Alarm (int MenuParam)
{	long int StartDelay;
	int AlarmTime;
	uint8_t i;

	for ( AlrmTmrNbr=1; AlrmTmrNbr<NBR_OF_ALARMS; AlrmTmrNbr++ )
		if ( !MODE.Setup ) break;			// if alarm not used
	if ( AlrmTmrNbr >= NBR_OF_ALARMS )
	{	LCD_Print ( "\nNo alarms available!");
		Delay ( 1000 );
		return 0;
	}
	LCD_Print ( "Using Alarm ");
	LCD_Char ( AlrmTmrNbr+'0' );
	if ( MenuParam == 1 )
	{	LCD_Print ( "\nEnter Delay Time:\n    ");
		StartDelay = 1;
		i = GetNumber( &StartDelay, 3, "Mins", 0 );
		if ( !i ) return -1;
		START = CurrentMins + StartDelay;
	}
	else if ( MenuParam == 2 )
	{	LCD_Print ( "\nEnter Alarm Time:\n    ");
		AlarmTime = Enter_Time (CurrentMins);
		if ( AlarmTime < 0 ) return -1;
		START = AlarmTime;
		if ( START < CurrentMins ) START += MPD; // set for tomorrow
	}
	else return -2;
	
	MODE.EZ_Timer = MODE.Enabled = MODE.On = MODE.Setup = 1;
	OUTPUTS.f.Spkr = 1;
	TONE = Sys.TMR_Tone;
	if ( START > MPD )					// if start time is tomorrow
	{	START -= MPD;
		DAYS = 1<<((DayOfWeek+1)%7);
	}
	else DAYS = 1<<DayOfWeek;			// set alarm for today only
	STOP = START + Sys.TMR_Stop;
	if ( STOP > MPD ) STOP -= MPD;		// if stop time is tomorrow
	Write_Alarm ( AlrmTmrNbr );

	return 1;
} // Mnu_1_Time_Alarm_Minutes


// get all the parameters needed for an alarm
int8_t Mnu_Full	(int MenuParam)
{	MenuParam &= ~1;					// do not ask for alarm nbr
	ActiveAlrmTmrNbr = Display_All_Alarms(); // get the alarm number
	if ( !ActiveAlrmTmrNbr ) return 0;
	AlrmTmrNbr = ActiveAlrmTmrNbr;		// copy for macros
	memset ( &MODE, 0, sizeof(MODE) );
	OUTPUTS.b = 0;
	DAYS = 0;
	TONE = 400/8;
	Mnu_Time (0);						// start and stop times
	Mnu_Days (0);						// days of week
	Mnu_Outputs (0);					// outputs
	if ( OUTPUTS.f.Spkr )
		Mnu_Tone(0);					// get speaker frequency
	Mnu_Repeat (0);
	Mnu_Enable (0);
	if ( MODE.Enabled && OUTPUTS.f.Spkr )
		MODE.On = 1;					// allow alarm to ring
	MODE.Setup = 1;						// show alarm/timer is setup
	Mnu_Save (1);
	return 1;
} // Mnu_Full


// display all the alarms
int8_t Mnu_Display (int MenuParam)
{
	Display_All_Alarms();
	return 1;
}


// get the start and stop times for the alarm/timer
// exit with: value stored in EE_Alarms
int8_t Mnu_Time (int MenuParam)
{	int StartTime, StopTime;

	LCD_Clear();
	if ( MenuParam & 1 )
		ActiveAlrmTmrNbr = Display_All_Alarms();
	if ( !ActiveAlrmTmrNbr ) return 0;
	AlrmTmrNbr = ActiveAlrmTmrNbr;		// copy for macros
// use times from normal array
	if ( !MODE.Cycle )					// if NOT cyclic timer
	{	if ( SETUP )					// get existing values
		{	StartTime = START;
			StopTime = STOP;
		}
		else
		{	StartTime = CurrentMins;
		}
	}
// use times from cycle array
	else
	{	StartTime = CycleTimes[AlrmTmrNbr-1][0];
		StopTime = CycleTimes[AlrmTmrNbr-1][1];
	}
// get values from operator
	LCD_Print ( "Enter Start Time:\n    ");
	StartTime = Enter_Time (StartTime);
	if ( !MODE.Cycle && !SETUP )
		StopTime = StartTime + EASY_ALARM_STOP;	
	if ( StopTime > MPD ) StopTime -= MPD;
	LCD_Print ( "\nEnter Stop Time:\n    ");
	StopTime = Enter_Time (StopTime);
// copy to EEPROM array
	START = StartTime;
	STOP = StopTime;
	if ( MenuParam & 1 ) MODE.Setup = 1, Mnu_Save(1);
	return 1;
} // Mnu_Time


// get the days of the week that the alarm/timer is to be enabled
// ActiveAlrmTmrNbr must already be set
// exit with: value stored in EE_Alarms
int8_t Mnu_Days	(int MenuParam)
{	uint8_t DaysOfWeek;

	LCD_Clear();
	if ( MenuParam & 1 )
		ActiveAlrmTmrNbr = Display_All_Alarms();
	if (!ActiveAlrmTmrNbr ) return 0;
	AlrmTmrNbr = ActiveAlrmTmrNbr;		// copy for macros
	DaysOfWeek = DAYS;
	LCD_Print ("Set Alarm/Timer Days\n");
	LCD_Print ( (char*)DaysShort );
	LCD_Char ( '\n' );
	DaysOfWeek = Select_N_on_Line ( DaysOfWeek, DaysShort_Cols );
	DAYS = DaysOfWeek;
	if ( MenuParam & 1 ) MODE.Setup = 1, Mnu_Save(1);
	return 1;
} // Mnu_Days


// exit with: value stored in EE_Alarms
int8_t Mnu_Outputs (int MenuParam)
{	uint8_t Outputs, i ,j;
	uint8_t Outputs_Cols[8];

	LCD_Clear();
	if ( MenuParam & 1 )
		ActiveAlrmTmrNbr = Display_All_Alarms();
	if ( !ActiveAlrmTmrNbr ) return 0;
	AlrmTmrNbr = ActiveAlrmTmrNbr;		// copy for macros
	LCD_Print ( "Select output(s):\nSpk ");
	j = (NBR_OF_COLS-4)/NBR_OUTPUTS;
	Outputs_Cols[0] = 2;				// LCD column for the speaker
	for ( i=0; i<NBR_OUTPUTS; i++ )		// for each output
	{	LCD_Pos ( 0, 5+(i*j) );
		LCD_Char ( i +'1' );
		Outputs_Cols[i+1] = 5+(i*j);
	}
	LCD_Char ('\n');
	Outputs_Cols[i+1] = 0;

	Outputs = OUTPUTS.f.Spkr;			// init. byte for outputs value
	if ( OUTPUTS.f.O1 ) Outputs |= BIT_1;
	if ( OUTPUTS.f.O2 ) Outputs |= BIT_2;
	if ( OUTPUTS.f.O3 ) Outputs |= BIT_3;
//jlc	if ( OUTPUTS.f.O4 ) Outputs |= BIT_4;
	Outputs = Select_N_on_Line ( Outputs, Outputs_Cols );
	OUTPUTS.f.Spkr = ( Outputs & BIT_0 );
	OUTPUTS.f.O1   = ( Outputs & BIT_1 ) ;
	OUTPUTS.f.O2   = ( Outputs & BIT_2 ) ;
	OUTPUTS.f.O3   = ( Outputs & BIT_3 ) ;
//jlc	OUTPUTS.f.O4   = ( Outputs & BIT_4 ) ;

	if ( MODE.Enabled && OUTPUTS.f.Spkr )
		MODE.On = 1;					// allow alarm to ring
	if ( MenuParam & 1 ) MODE.Setup = 1, Mnu_Save(1);
	return 1;
} // Mnu_Outputs


// exit with: value stored in EE_Alarms
int8_t Mnu_Tone (int MenuParam)
{
	LCD_Clear();
	if ( MenuParam & 1 )
		ActiveAlrmTmrNbr = Display_All_Alarms();
	if ( !ActiveAlrmTmrNbr ) return 0;
	AlrmTmrNbr = ActiveAlrmTmrNbr;		// copy for macros
	LCD_Print ("Enter Alarm Tone\nFrequency:");
	ToneFrequency = (long int)TONE * 8;
	if (!ToneFrequency) ToneFrequency = 400;
	Set_Spkr_Freq (ToneFrequency);
	OC3_Start();
	GetNumber ( &ToneFrequency, 4, "Hz", &Set_Spkr_Freq );
	OC3_Stop();
	TONE = ToneFrequency/8;
	if ( MenuParam & 1 ) MODE.Setup = 1, Mnu_Save(1);
	return 1;
} // Mnu_Tone


// set the speaker frequency
// entry argument: speaker frequency in Hz
// the long int argument is for compatibility with GetNumber and its callback feature
void Set_Spkr_Freq ( long int TF )
{	int i,y;

	y = TF;					// for debugging
	i = y % 8;
	if (i)
	{	if ( i >  4) y = ((y/8) * 8) - 8;
		else y = ((y/8) * 8) + 8;
	}
	ToneFrequency = y;
	y = _XTAL_FREQ/2/TF;
	OC3RS = y;
	OC3R =  y/2;
} // Set_Spkr_Freq


// set the on and off times for the cyclic option
// IMPORTANT: the selected alarm/timer must already be set to the beginning
// and ending times of the cycle
// Both the start and stop times will be divided by CYCLE_RESOLUTION and then
//		stored.  Therefore, they will actually be decreased by up to
//		(CYCLE_RESOLUTION-1) minutes.
// exit with: value stored in EE_Alarms
int8_t Mnu_Cycle (int MenuParam)
{	int CyclePeriod, CycleSpan, i;

	LCD_Clear();
	if ( MenuParam & 1 )
		ActiveAlrmTmrNbr = Display_All_Alarms();
	if ( !ActiveAlrmTmrNbr ) return 0;
	AlrmTmrNbr = ActiveAlrmTmrNbr;		// copy for macros
	if ( !MODE.Setup )					// if the alarm has not been set up
	{	LCD_Print ( "This Alarm/Timer\nhas not been setup!\nPress Sw3 to exit");
		WaitPB (SW_3);
		return 0;
	}
	MODE.Cycle = 1;
	LCD_Print ("    Cyclic Timer\nMins On:" );
	CycleMinutes = DEFAULT_CYCLE_TIME;
	if ( !GetNumber ( &CycleMinutes, 3, "", 0 ) ) return 0;;
	CYCLE_ON = CycleMinutes/CYCLE_RESOLUTION;
	LCD_Print ( " Off:" );
	CycleMinutes = DEFAULT_CYCLE_TIME;
	if ( !GetNumber ( &CycleMinutes, 3, "", 0 ) ) return 0;;
	CYCLE_OFF = CycleMinutes/CYCLE_RESOLUTION;
// calculate end of span: off time of last cycle
	CyclePeriod = (CYCLE_ON + CYCLE_OFF) * CYCLE_RESOLUTION;
	CycleSpan = STOP - START;
	if ( CycleSpan < 0 ) CycleSpan += MPD; // adjust for spanning midnight
	i = CycleSpan/CyclePeriod;			// nbr of cycles in span
	STOP = START + (i*CyclePeriod) + CYCLE_ON;
// save values
	CycleTimes[AlrmTmrNbr-1][0] = START; // save start and end times
	CycleTimes[AlrmTmrNbr-1][1] = STOP;
	if ( MenuParam & 1 )
	{	MODE.Cycle = 1;
		LCD_Char ('\n' );
		Mnu_Save(0);
	}
	return 1;
} // Mnu_Cycle


// exit with: value stored in EE_Alarms
int8_t Mnu_Enable (int MenuParam)
{
	LCD_Clear();
	if ( MenuParam & 1 )
		ActiveAlrmTmrNbr = Display_All_Alarms();
	if ( !ActiveAlrmTmrNbr ) return 0;
	AlrmTmrNbr = ActiveAlrmTmrNbr;		// copy for EE_ macros
	LCD_Print ( "Enable? " );
	if ( Yes_No() )
	{	MODE.Enabled = 1;
		if (OUTPUTS.f.Spkr) MODE.On = 1; // allow alarm to ring
		else MODE.On = 0;				// disable ringing
	}
	else
		MODE.Enabled = MODE.On = 0;		// disable ringing

	if ( MenuParam & 1 ) MODE.Setup = 1, Mnu_Save(1);
	return 1;
} // Mnu_Enable


// exit with: value stored in EE_Alarms
int8_t Mnu_Repeat (int MenuParam)
{
	LCD_Clear();
	if ( MenuParam & 1 )
		ActiveAlrmTmrNbr = Display_All_Alarms();
	if ( !ActiveAlrmTmrNbr ) return 0;
	AlrmTmrNbr = ActiveAlrmTmrNbr;		// copy for macros
	LCD_Print ( "Repeat Weekly?\n   " );
	if ( Yes_No() )
		MODE.Repeat = 1;
	else
		MODE.Repeat = 0;
	if ( MenuParam & 1 ) MODE.Setup = 1, Mnu_Save(1);
	return 1;
} // Mnu_Repeat


// exit with: EE_Alarms values cleared if Yes
int8_t Mnu_Clear (int MenuParam)
{
	LCD_Clear();
	if ( MenuParam & 1 )
		ActiveAlrmTmrNbr = Display_All_Alarms();
	if ( ActiveAlrmTmrNbr <= 0 ) return 0;
	Display_Alarm ( ActiveAlrmTmrNbr );
	LCD_Print ( "Clear? " );
	if ( Yes_No() )
	{	memset ( &Alarms[ActiveAlrmTmrNbr-1], 0, bALARM_SIZE );
		Write_Alarm ( ActiveAlrmTmrNbr );
	}
	return 1;
} // Mnu_Clear


// exit with: values written to EEPROM if Yes
int8_t Mnu_Save (int MenuParam)
{
	if ( MenuParam ) Display_Alarm ( ActiveAlrmTmrNbr );
	LCD_Print ( "\nSave? " );
	if ( Yes_No() ) Write_Alarm ( ActiveAlrmTmrNbr );
	return 1;
} // Mnu_Save


/*******************************************************************************
						Alarm/Timer Utility Functions
*******************************************************************************/

// display the values for all alarms/timers
// return value: the selected alarm
//				0 = Sw3 exit, no alarm chosen - abort
// the alarm data is from EEPROM and stored into EE_Alarms[AlrmTmrNbr-1]
uint8_t	 Display_All_Alarms ( void )
{	uint8_t SwValue, i;

	i = 1;
	while (1)
	{	Display_Alarm ( i );
		SwValue = WaitSwitches();		// wait for a switch
		if ( SwValue == SW_QE ) break;
		if ( SwValue == CW )
		{	i++;
			if ( i > NBR_OF_ALARMS ) i = 1;
		}
		else if ( SwValue == CCW )
		{	i--;
			if ( i < 1 ) i = NBR_OF_ALARMS;
		}
		else if ( SwValue == SW_3 ) return 0;
	}
	LCD_Clear();
	return i;
} //Display_All_Alarms


// display the data for an alarm/timer
// entry argument: alarm/timer number
// the data is read from EE_Alarms
void Display_Alarm ( int8_t AlrmTmrNbr )
{	uint8_t DaysOfWeek, i;
	int Tone;

	LCD_Clear();
	if ( MODE.Setup )					// if alarm/timer has been setup
	{
// line 1: Alarm nbr, outputs, enable
		LCD_Char ( AlrmTmrNbr+'0' );	// display alarm number
		if ( MODE.Repeat ) LCD_Char ( 'R' ); // show repeat mode
		else LCD_Print ("NR");			// show not repeat
		if ( OUTPUTS.f.Spkr )
		{	Tone = (int)TONE * 8;
			Nbr_to_ASCII ( Tone, 4, &LCDstring[0], " Spk ", 0 );
			LCD_Print ( LCDstring );
		}
		LCD_Print ( " O:");
		if ( OUTPUTS.f.O1 ) LCD_Char ('1');
		if ( OUTPUTS.f.O2 ) LCD_Char ('2');
		if ( OUTPUTS.f.O3 ) LCD_Char ('3');
		if ( OUTPUTS.f.O4 ) LCD_Char ('4');
		LCD_Pos ( 0, 18 );
		if ( MODE.Enabled ) LCD_Print (" En");
		else LCD_Print ("-En");
		LCD_Char ( '\n' );
// Line 2: active days
		DaysOfWeek = DAYS;
		if (DaysOfWeek)
		{	for ( i=0; i<=6; i++ )		// for each day
			{	if ( DaysOfWeek & (1<<i) ) // if day is enabled
					LCD_PrintN ( (char*)&DaysShort[i*3], 3 );
			}
			LCD_Char ( '\n' );
		}
// Line 3: start and stop times
		if ( !MODE.Cycle )				// if NOT cyclic timer
		{	Display_Time ( START, 0 );
			LCD_Print ("  ");
			Display_Time ( STOP, 0 );
			LCD_Char ( '\n' );			// position to line 4
		}
		else
		{	Display_Time ( CycleTimes[AlrmTmrNbr-1][0], 0 );
			LCD_Print ("  ");
			Display_Time ( CycleTimes[AlrmTmrNbr-1][1], 0 );
			LCD_Char ( '\n' );			// position to line 4
	// Line 4: cycle on/off times
			i = Nbr_to_ASCII ( (long)CYCLE_ON*CYCLE_RESOLUTION, 3, LCDstring, "Cycle: ", 0 );
			Nbr_to_ASCII ( (long)CYCLE_OFF*CYCLE_RESOLUTION, 3, &LCDstring[i], " ", 0 );
			LCD_Print (LCDstring);
			Delay ( 2000 );
		}
	}
	else // if alarm/timer has NOT been setup
	{	LCD_Print ( "Alarm/Timer " );
		LCD_Char ( AlrmTmrNbr+'0' );	// display alarm number
		LCD_Print ( "\n  available\n" );
	}
} // Display_Alarm


/*******************************************************************************
							Utility Functions
*******************************************************************************/

int8_t Mnu_Snooze (int MenuParam)
{	long int i;

	i = SnoozeTime = Sys.Snooze;		// get current snooze value
	LCD_Print ("Enter Snooze time\n   ");
	if ( !GetNumber ( &SnoozeTime, 2, " Minutes", 0 ) ) return 0; // get new value
	if ( i != SnoozeTime )				// if different
	{	if ( SnoozeTime > MAX_SNOOZE ) SnoozeTime = MAX_SNOOZE;
		Sys.Snooze = SnoozeTime;
		Write_Alarm(0);
	}
	return 1;
} // Mnu_Snooze


int8_t Mnu_LCD_Bright (int MenuParam)
{	long int temp;

	LCD_Print ("AM time: ");
	Sys.Bright_tm_day = Enter_Time (Sys.Bright_tm_day);
	LCD_Print ("\n  Brightness: ");
	temp = Sys.Bright_Day;
	if ( temp > 100  ||  temp < 1 )temp = 50;
	GetNumber ( &temp, 2, "%", &Set_LCD_Brightness );
	Sys.Bright_Day = temp;

	LCD_Print ("\nPM time: ");
	Sys.Bright_tm_nite = Enter_Time (Sys.Bright_tm_nite);
	LCD_Print ("\n  Brightness: ");
	temp = Sys.Bright_Nite;
	if ( temp > 100  ||  temp < 1 )temp = 50;
	GetNumber ( &temp, 2, "%", &Set_LCD_Brightness );
	Sys.Bright_Nite = temp;

	Write_Alarm(0);
	LCD_Bright_Day_Night ();

	return 1;
} // Mnu_LCD_Bright


// set the LCD brightness by varying the duty cycle
// entry argument: brightness level as a %
// the long int argument is for compatibility with GetNumber and its callback feature
void Set_LCD_Brightness( long int i )
{	float y;

	if ( i > 99 ) i = 99;
	if ( i < 1 ) i = 10;
	y = ((float)i/100.0) * (float)LCD_PERIOD;
	OC2R = (uint16_t)y;
} // Set_LCD_Brightness


// get new timer divisor from operator
// exit with: value stored in system memory (Sys) if Yes
int8_t Mnu_Timer_Div (int MenuParam)
{	long int Divisor;

	Divisor = PR3;
	Divisor <<= 16;
	Divisor += PR2;
	Nbr_to_ASCII ( Divisor, 8, &LCDstring[0], "Divisor: ", 0 );

	LCD_Print (LCDstring);
	LCD_Print ( "\nChange it: ");
	if ( Yes_No() )
	{	LCD_Print ( "\nNew Value:" );
		GetNumber ( &Divisor, 8, "", 0 );
		Sys.Tmr2_Divlo = Divisor;
		Sys.Tmr2_Divhi = Divisor>>16;
		Start_Timer(1);					// saves to EEPROM
	}
	return 1;
} // Mnu_Timer_Div


// read and display battery and main power voltages
// option to clear the Power Off counter
int8_t Mnu_Power (int MenuParam)
{	uint8_t FlagTemp, Count, j;

	Nbr_to_ASCII ( (long int) PowerOffCount, 2, LCDstring, "Power Failures: ", "\n" );
	LCD_Print (LCDstring);
	if ( PowerOffCount )
	{	LCD_Print ( "\4SW2 to continue" );
		FlagTemp = Sys.Flag.Hr24;		// save AM/PM flag
		Sys.Flag.Hr24 = 1;				// force 24hr display format
		Count = PowerOffCount;
		if ( Count > MAX_POWER_CYCLES ) Count = MAX_POWER_CYCLES;
		for ( j=0; j<Count; j++ )
		{	LCD_ClearLine(2);
			Display_Time ( Power[j].OffTime, 0 );
			Nbr_to_ASCII ( (long int) Power[j].OffDay, 2, LCDstring, " ", "   " );
			LCD_Print (LCDstring);		// show day of month
			if ( Power[j].OnTime > 0 )
			{	Display_Time ( Power[j].OnTime, 0 );
				Nbr_to_ASCII ( (long int) Power[j].OnDay, 2, LCDstring, " ", 0 );
				LCD_Print (LCDstring);	// show day of month
			}
			WaitPB (SW_2);
			Delay(500);
		}
		Sys.Flag.Hr24 = FlagTemp;		// restore AM/PM flag
		LCD_Print ("\4Clr Failures: ");
		if ( Yes_No() ) PowerOffCount = 0;
	}
	ShowVoltages();
	LCD_ClearLine(4);
	LCD_Print ( "SW3 to exit" );
	WaitPB (SW_3);

	return 1;
} // Mnu_Power


// enter with cursor on display line
void ShowVoltages ( void )
{	double Vbatt, Vin;

	Vbatt = ReadAD(AD_BATT);
	Vin = ReadAD(AD_VIN);
	LCD_ClearLine(0);
	LCD_Print ("Vin=      VBatt=    ");
	FP2Ascii ( Vin, 5 );
	FP2Ascii ( Vbatt, 17 );
} // ShowVoltages


// convert a double to ASCII in X.1%d format
// enter with:
//		the number to convert
//		the offset on the LCD to put the ASCII value
// exit with: the ASCII value displayed
void FP2Ascii ( double V, uint8_t i )
{	int j, k;
	char MyString[10];

	LCD_Pos(0,i);
	j = (V*10.0);
// the string builds starting with the tenths digit
	k = 0;
	MyString[k++] = (j%10) + '0';		// tenths digit
	MyString[k++] = '.';
	while (j/10)
	{	j /= 10;
		MyString[k++] = (j%10) + '0';
	}
	while (k>0) LCD_Char ( MyString[--k] );
} // FP2Ascii


int8_t Mnu_DST (int MenuParam)
{	uint8_t flags;

	LCD_Clear();
	flags = Sys.Flag.DST_En;
	LCD_Print ( "DST detection is " );
	if (Sys.Flag.DST_En) LCD_Print ("On");
	else LCD_Print ("Off");
	LCD_Print ( "\nEnable ? ");
	if ( Yes_No() ) Sys.Flag.DST_En = 1;
	else Sys.Flag.DST_En = 0;

	if ( Sys.Flag.DST_En )				// if DST detection is enabled
	{	LCD_Print ( "\nDST is " );
		if ( Sys.Flag.DST ) LCD_Print ("On");
		else LCD_Print ("Off");
		LCD_Print ( "\nSet to DST? ");
		if ( Yes_No() ) Sys.Flag.DST = 1;
		else Sys.Flag.DST = 0;
	}
	if ( flags != Sys.Flag.DST_En ) Write_Alarm (0);
	return 1;
} // Mnu_DST


int8_t Mnu_Disp_Secs (int MenuParam)
{	uint8_t i;

	if ( !Sys.Flag.Hr24 ) TimeLength = 8;
	else TimeLength = 5;
	if ( Sys.Flag.Secs )
		LCD_Print ( "Seconds display: On" );
	else
		LCD_Print ( "Seconds display: Off" );
	i = Sys.Flag.Secs;
	LCD_Print ( "\nShow seconds? " );
	if ( Yes_No() )
		TimeLength += 3, Sys.Flag.Secs = 1;
	else
		Sys.Flag.Secs = 0;
	if ( i != Sys.Flag.Secs ) Write_Alarm (0);
	return 1;
} // Mnu_Disp_Secs


const uint8_t DateFormat[] = {3,10,0};
int8_t Mnu_Date_Format ( int MenuParam )
{	uint8_t i;
	LCD_Print ("\rDate format is ");
	if ( !Sys.Flag.Format ) LCD_Print ("M/D/Y");
	else LCD_Print ("D/M/Y");
	LCD_Print ("\nM/D/Y  D/M/Y");
//				  1234567890
	i = Select_1_on_Line ( DateFormat );
	if ( i<0 ) return -1;
	if ( i ) Sys.Flag.Format = 1;
	else Sys.Flag.Format = 0;
	return 1;
} // Mnu_Date_Format

int8_t Mnu_EZ_Al_Days	(int MenuParam)
{	uint8_t ez_byte;
	LCD_Print ("Easy Alarm Days\n");
	LCD_Print ( (char*)DaysShort );
	LCD_Char ( '\n' );
	ez_byte = Sys.EZ_Days;				// get current value
	ez_byte = Select_N_on_Line ( ez_byte, DaysShort_Cols );
	Sys.EZ_Days = ez_byte;
	return 1;
}

int8_t Mnu_EZ_Al_Tone	(int MenuParam)
{	long int ez_value;
	LCD_Clear();
	LCD_Print ("Easy Tone: ");
	ez_value = (long)Sys.EZ_Tone * 8;
	GetNumber ( &ez_value, 4, " Hz", 0 );
	Sys.EZ_Tone = ez_value/8;
	return 1;
}

int8_t Mnu_EZ_Al_Stop	(int MenuParam)
{	long int ez_value;
	LCD_Clear();
	LCD_Print ("Easy Stop: ");
	ez_value = (long)Sys.EZ_Stop;
	GetNumber ( &ez_value, 3, " mins", 0 );
	Sys.EZ_Stop = ez_value;
	return 1;
}

int8_t Mnu_EZ_Al_Rpt	(int MenuParam)
{	uint8_t ez_byte;
	LCD_Clear();
	LCD_Print ("Easy Repeat: ");
	ez_byte = Yes_No();
	Sys.EZ_Repeat = ez_byte;
	return 1;
}

int8_t Mnu_EZ_Tmr_Stop	(int MenuParam)
{	long int ez_value;
	LCD_Clear();
	LCD_Print ("Easy Timer Stop:\n    ");
	ez_value = (long)Sys.TMR_Stop;
	GetNumber ( &ez_value, 3, " mins", 0 );
	Sys.TMR_Stop = ez_value;
	return 1;
}

int8_t Mnu_EZ_Tmr_Tone	(int MenuParam)
{	long int ez_value;
	LCD_Clear();
	LCD_Print ("Easy Timer Tone:\n     ");
	ez_value = (long)Sys.TMR_Tone * 8;
	GetNumber ( &ez_value, 4, " Hz", 0 );
	Sys.TMR_Tone = ez_value/8;
	return 1;
}

int8_t Mnu_EZ_Save (int MenuParam)
{	LCD_Print ( "\nSave Easy values?\n   " );
	if ( Yes_No() ) Write_Alarm (0);
	return 1;
} // Mnu_Easy_Values


int8_t Mnu_Info (int MenuParam)
{	LCD_Clear();
	LCD_Print ("Clock/Timer by K3PTO\n Built: ");
	LCD_Print (__DATE__);
	LCD_Char ('\n');
	ShowVoltages();
	LCD_Print ("\nSw3 to exit");
	WaitPB(SW_3);
	return 1;
} // Mnu_Info


// check battery and main power voltages
// enter with flag value:
//		bit 0 = 1 = check battery
//		bit 1 = 1 = check main power and log power failure times
// Battery: display state
// Main power: log power transitions
void CheckPower (uint8_t flag)
{	double fVolts;
	uint8_t i;

// check battery
	if ( flag & 1 )
	{	fVolts = ReadAD(AD_BATT);
		LCD_Print ( "Batt " );
		if ( fVolts >= MIN_BATT )
			LCD_Print ( "OK" ), Flags &= ~FLAG_LOW_BATT; // clear battery low flag
		else
			LCD_Print ( "Low" ), Flags |= FLAG_LOW_BATT; // set battery low flag
	}
// check for main power
	if ( flag & 2 )
	{	fVolts = ReadAD(AD_VIN);
		i = PowerOffCount % MAX_POWER_CYCLES; // allows maximum of MAX_POWER_CYCLES
		if ( fVolts >= MIN_VIN )		// if main power is OK
		{	if ( Flags & FLAG_PWR_FAIL ) // if power was out & is now OK
			{	Flags &= ~FLAG_PWR_FAIL; // clear the error flag
				Power[i-1].OnTime = CurrentMins; // save time
				Power[i-1].OnDay = DayOfMonth; //		and day
			}
			LCD_Bright_Day_Night ();
		}
		else							// if main power is down
		{	if ( Flags & FLAG_PWR_FAIL ) return; // if already flagged
			Flags |= FLAG_PWR_FAIL;		// show main power failure
			Power[i].OffTime = CurrentMins; // save time
			Power[i].OffDay = DayOfMonth; //		and day
			Power[i].OnTime = -1;		// show no on time yet
			Set_LCD_Brightness (Sys.Bright_Nite);
			PowerOffCount++;			// update power fail count
		}
	}
} // CheckPower


// Read an A/D channel
// entry argument: channel nbr: AD_VIN or AD_BATT
// return value: measured voltage
double ReadAD (uint8_t Channel)
{	double fVolts;
	uint16_t iVolts;

	AD1CHS = Channel;
	AD1CON1bits.SAMP = 1;				// enable S&H
	Delay (2);							// allow S&H to acquire signal
	AD1CON1bits.SAMP = 0;				// disable S&H and start conversion
	while ( AD1CON1bits.DONE == 0 );	// wait for conversion to complete
	iVolts = (ADC1BUF0 & 0x3FF);		// read the buffer, keep only bits 9..0
	fVolts = 5.0 * (double)iVolts/1024.0;
	fVolts *= AD_MULT;					// adjust for voltage divider
	return fVolts;
} // ReadAD


// Operation:
//		stop the timer
//		reset globals
//		set timer divisor
//		start timer
// entry argument: flag indicating execution options
//		1 = do all
//		0 = skip some operations
void Start_Timer(uint8_t flag)
{
	TMR2_Stop();						// ensure timer is stopped
	if ( flag )
	{	Seconds = 0;					// start at beginning of minute
		NewMinute = 0;
		CalMins = 0;					// minutes into calibration period
	}
	PR2 = Sys.Tmr2_Divlo;				// divisor LS word
	PR3 = Sys.Tmr2_Divhi;				//		and MS word
	TMR3HLD = 0;						// reset Timer
	TMR2 = 0;							//		counter
	TMR2_Start();
	Write_Alarm(0);						//  save the new value
} // Start_Timer


// Update time values when a new minute occurs:
//		global variables: CurrentMins, DayOfWeek, DayOfMonth, Month, Year
// Process other timed events
void Handle_New_Minute(void)
{
	NewMinute--;						// update ISR minute counter
	CurrentMins++;						// update minutes into day
	CalMins++;							// minutes into calibration period
	if ( CurrentMins >= MPD )			// if next day
	{	CurrentMins = 0;
		DayOfMonth++;
		if ( DayOfMonth > DaysPerMonth[Month-1] ) // if next month
		{	if ( Year%4!=0 || Month!=2 || DayOfMonth!=29 ) // if not leap year Feb 29
			{	DayOfMonth = 1;
				Month++;
				if ( Month > 12 )
				{	Month = 1;
					Year++;
				} // new year
			} // not  leap year Feb 29
		} // new month
		DayOfWeek++;
		if ( DayOfWeek >= 7 ) DayOfWeek = 0; // set to Sunday
	} // new day

// check for DST
	if ( (CurrentMins == 120)  &&  (DayOfWeek == 0) ) // if 2:00 AM on Sunday
	{	if ( Sys.Flag.DST_En )			// if DST detection enabled
		{	if ( Month == 3 )			// DST starts on second Sunday in March
			{	if ( DayOfMonth >= 8  &&  DayOfMonth <= 14 )
				{	CurrentMins = 180;	// set to 3:00
					Sys.Flag.DST = 1;	//  show DST in effect
					Write_Alarm (0);
				}
			} // if Month = 3
			if ( Month == 11 )			// and ends on first Sunday in November
			{	if ( DayOfMonth >= 1  &&  DayOfMonth <= 7 )
				{	if ( Sys.Flag.DST )	// if DST in effect
					{	CurrentMins = 60; // set to 1:00
						Sys.Flag.DST = 0; // show DST NOT in effect
						Write_Alarm (0);
					}
				}
			} // if Month =11
		} // if Sunday && DST detection enabled
	} // if 2:00 AM

// other timed events
	if ( !(Flags & FLAG_PWR_FAIL) ) LCD_Bright_Day_Night ();
} // Handle_New_Minute


/******************************************************************************/
// check to see if any alarm/timer activity required
// this function activates/deactivates the outputs
/******************************************************************************/
void Check_Alarms (void)
{	uint8_t i;

	for ( AlrmTmrNbr = 1; AlrmTmrNbr <= NBR_OF_ALARMS; AlrmTmrNbr++ )
	{	if ( MODE.Enabled )
		{	if ( START == CurrentMins )
			{	if ( DAYS & (1<<DayOfWeek) )
				{	if ( OUTPUTS.f.O1 ) Output_1_SetHigh();
					if ( OUTPUTS.f.O2 ) Output_2_SetHigh();
					if ( OUTPUTS.f.O3 ) Output_3_SetHigh();
// jlc					if ( Outputs & O4_ACTIVE ) Output_4_SetHigh();
					if ( OUTPUTS.f.Spkr )
					{	// program gets here IFF for this alarm:
						//		enabled AND time AND day AND speaker
						if ( MODE.On && (AlarmFlags & AF_ALARMS_EN) )
						{	Flags |= FLAG_RINGING; // turn on ringing via tmr2 ISR
							SnoozeAlarmNbr = AlrmTmrNbr; // in case snooze is requested
						}
						AlarmFlags |= AF_ALARMS_EN; // force alarms enabled
						if ( DAYS ) MODE.On = 1;// allow alarm to ring
						else MODE.On = 0; // disable future ringing
					} // SPKR_ACTIVE
					if ( MODE.Cycle ) // if cyclic timer
						STOP = START + (CYCLE_ON*CYCLE_RESOLUTION);
				} // this day of week
			} // start time
	// stop time
			// putting the stop time test here allows the stop time
			//	to occur in the next day
			if ( STOP == CurrentMins )
			{	if ( OUTPUTS.f.Spkr )	// &&  (AlrmTmrNbr != SnoozeAlarmNbr)
				{	Flags &= ~FLAG_RINGING; // tell ISR
					OC3_Stop();			// turn off the speaker
					if ( DAYS ) MODE.On = 1; // allow alarm to ring
					else MODE.On = 0;	// disable ringing
				}
				if ( OUTPUTS.f.O1 ) Output_1_SetLow();
				if ( OUTPUTS.f.O2 ) Output_2_SetLow();
				if ( OUTPUTS.f.O3 ) Output_3_SetLow();
//jlc				if ( OUTPUTS.f.O4 ) Output_4_SetLow();
				i = (MODE.Repeat); // =1 if repeating
				if ( MODE.Cycle ) // if cyclic timer
				{	if ( STOP == CycleTimes[AlrmTmrNbr-1][1] ) // if end of span
						START = CycleTimes[AlrmTmrNbr-1][0]; // start of next span
					else
						i=1, START = STOP + (CYCLE_OFF*CYCLE_RESOLUTION);
							// do not clear MODE.Enabled if not end of span
				}
				if ( !i )				// if NOT repeating
				{	DAYS &= ~(1<<DayOfWeek); // remove from schedule
					if ( !DAYS ) MODE.Enabled = 0; // disable if no more days
					Write_Alarm(AlrmTmrNbr);
				}
				if ( MODE.EZ_Timer ) memset ( &MODE, 0, sizeof(MODE) );
			} // stop time
		} // alarm enabled
	} // for each alarm

	if ( AlarmFlags & AF_SNOOZE )		// if alarm is snoozing
	{	SnoozeTime--;					// update time remaining
		if ( !SnoozeTime )				// if done snooze time
		{	AlrmTmrNbr = SnoozeAlarmNbr; // needed by MODE
			if ( MODE.On )
			{	Flags |= FLAG_RINGING;	// turn on ringing via tmr2 ISR
				AlarmFlags &= ~AF_SNOOZE; // snooze time is done
			}
		}
	}

} // Check_Alarms


/*******************************************************************************
						Time and Date functions
*******************************************************************************/

// display a time value
// entry arguments:
//		minutes past midnight
//		flag: 1 = display seconds if Sys.Flag.Secs is set
// IMPORTANT: the screen and cursor must be set before calling this function
//		If using AM/PM there must be at least 8 columns remaining on the LCD line
//		If using 24 hours, there must be at least 5 columns
//		If displaying seconds, 3 more columns are needed
const char AMPM[] = { " AM PM" };

void Display_Time ( int Time, uint8_t flag )
{	uint8_t hour, minute, ampm, j;

	hour = Time/60;						// get hour
	minute = Time % 60;					//		and minute values
	if ( !Sys.Flag.Hr24 )				// if using AM/PM
	{	if ( hour > 12 ) hour -= 12;
		if ( hour == 0 ) hour = 12;
		if ( Time >= NOON ) ampm = 3;	// index into AMPM
		else ampm = 0;
	}
	j = Nbr_to_ASCII ( hour, 2, &LCDstring[0], 0, ":" );
	j += Nbr_to_ASCII ( minute, -2, &LCDstring[j], 0, 0 );

	if ( flag )							// if print seconds
	{	if ( Sys.Flag.Secs )
			j += Nbr_to_ASCII ( Seconds, -2, &LCDstring[j], ":", 0 );
	}
	LCD_Print ( LCDstring );
	if ( !Sys.Flag.Hr24 )				// if using AM/PM
		LCD_PrintN ( (char*)&AMPM[ampm], 3 );
} // Display_Time


// Operator to enter a time value
// entry argument: initial time value in minutes past midnight
//		LCD screen must be set as desired
// return value:
//		normal: the entered time value in minutes past midnight
//		SW3 exit: -1
#define OFF_HOUR	1		// hh:mm
#define OFF_MINUTE	4		// 01234
int Enter_Time ( int Time )
{	uint8_t SwValue, offset, column;

	if ( Time > MPD  ||  Time < 0 ) Time = MPD/2;
	column = LCD_Column();				// get starting column number
	offset = OFF_HOUR;					// column offset for value being modified
	while (1)
	{	LCD_Pos ( 0, column );
		Display_Time ( Time, 0 );
		LCD_Pos ( 0, column+offset );
		SwValue = WaitSwitches();		// wait for a switch
		if (SwValue == CW)
		{	if ( offset == OFF_HOUR )	// update hours value
			{	Time += 60;
				if ( Time >= MPD ) Time -= MPD;
			}
			else						// must be minutes
			{	Time++;
				if ( Time >= MPD ) Time = 0;
				if ( Time%60 == 0 ) Time -= 60; // if next hour, backup one hour
			}
		}
		if (SwValue == CCW)				// hour and minute are UNSIGNED
		{	if ( offset == OFF_HOUR )	// update hours value
			{	Time -= 60;
				if ( Time < 0 ) Time += MPD;
			}
			else						// must be minutes
			{	Time--;
				if ( Time < 0 ) Time = MPD-1;
				if ( Time%60 == 59 ) Time += 60; // if previous hour, advance 1 hour
			}
		}
		if (SwValue == SW_QE)			// accept the values
		{	if ( offset == OFF_HOUR ) offset = OFF_MINUTE;
			else break;
		}
		if ( SwValue == SW_2 )			// backup to hour
			if ( offset == OFF_MINUTE ) offset = OFF_HOUR;
		if ( SwValue == SW_3 )			// cancel
			return(-1);
	}
	return Time;
} // Enter_Time


// display the date
// enter with the cursor on the line to be used
void Display_Date (void)
{	char DateString[25];
	uint8_t i,j;

	if ( DayOfMonth < 10 ) LCD_Pos (0,2);
	else  LCD_Pos (0,1);
	i = j = 0;
// start with day name (includes comma))
	while ( DayNames[DayOfWeek][i] )
		DateString[i++] = DayNames[DayOfWeek][j++];
	DateString[i++] = ' ';

	if ( !Sys.Flag.Format )				// insert month and day
	{		i += Nbr_to_ASCII ( DayOfMonth, 2, &DateString[i], (char*)&MonthNames[Month-1], "," );
	}
	else								// insert day and month
	{	i += Nbr_to_ASCII ( DayOfMonth, 2, &DateString[i], 0, " " );
		j = 0;
		while ( MonthNames[Month-1][j] )
			DateString[i++] = MonthNames[Month-1][j++];
		DateString[--i] = ',', i++;		// replace space with comma
	}
// insert year
	Nbr_to_ASCII ( Year, 4, &DateString[i], " ", 0 );
	LCD_Print ( DateString );
} // Display_Date


void LCD_Bright_Day_Night ( void )
{	if ( (CurrentMins >= Sys.Bright_tm_day)  &&  (CurrentMins < Sys.Bright_tm_nite) )
		Set_LCD_Brightness( Sys.Bright_Day );
	else
		Set_LCD_Brightness( Sys.Bright_Nite );
}

/*******************************************************************************
						EEPROM  functions
*******************************************************************************/
/* EEPROM is at the top of program flash and is addressable as 16 bit words.
 The memory is byte addressable although writes can only be performed on word
 addresses.  Each accessible word starts on on even address.  The offset value
 for these functions is the WORD offset.  The functions multiply the offset by 2
 in order to get the byte address.
*/


// read the alarm/timer data from EEPROM
// entry argument:
//		alarm/timer number: 1..7
//			0 = System Data
// the data will be stored in structure EE_Alarms or system memory (Sys)
//
// the __builtin_tblrdl requires:
//		setting TBLPAG to the memory page being accessed
//		using a byte offset into the page for the read
void Read_Alarm ( int8_t AlrmTmrNbr )
{	unsigned int EE_Offset;
	uint8_t i,j;
	uint16_t *pointer;

	TBLPAG = EE_PAGE;
	EE_Offset = EE_OFFSET;
	if ( AlrmTmrNbr >= 1 )				// do an alarm
	{	EE_Offset += bSYS_SIZE;			// adjust for system data
		j = AlrmTmrNbr-1;
		EE_Offset += j*bALARM_SIZE;		// offset to this alarm
		pointer = &Alarms[AlrmTmrNbr-1].Start;
		for ( i=0; i<wALARM_SIZE; i++)	// for each word of alarm data
			*pointer++ = __builtin_tblrdl(EE_Offset+(i*2));
		if ( Alarms[AlrmTmrNbr-1].Start == 0xFFFF ) // if EEPROM is erased
			memset ( &Alarms[AlrmTmrNbr-1].Start, 0, bALARM_SIZE ); // clear the array
	}
	else								// do system blocks
	{	pointer = &Sys.Tmr2_Divlo;
		for ( i=0; i<wSYS_SIZE; i++)	// for each word of system data
			*pointer++ = __builtin_tblrdl(EE_Offset+(i*2));
	}
} // Read_Alarm


// write the alarm/timer data to EEPROM
// entry argument:
//		alarm/timer number: 1..7
//			0 = System Data
// the data must be read from structure EE_Alarms or system memory (Sys)
// the _write_eedata_word function requires the full EEPROM address
void Write_Alarm ( int8_t AlrmTmrNbr )
{	_prog_addressT EE_Addr;
	uint8_t i,j;
	uint16_t *pointer;

	EE_Addr = EE_BASE;					// base addr of EEPROM
	if ( AlrmTmrNbr >= 1 )				// do an alarm
	{	EE_Addr += bSYS_SIZE;			// adjust for system data
		j = AlrmTmrNbr-1;
		EE_Addr += j*bALARM_SIZE;		// address of this alarm
		pointer = &Alarms[AlrmTmrNbr-1].Start;
		for ( i=0; i<wALARM_SIZE; i++)	// for each word of alarm data
		{	_write_eedata_word( EE_Addr, *pointer++ );
			EE_Addr += 2;				// address of next word in EEPROM
			_wait_eedata();
		}
	}
	else								// do system data
	{	pointer = &Sys.Tmr2_Divlo;
		for ( i=0; i<wSYS_SIZE; i++ )	// for each word of system data
		{	_write_eedata_word( EE_Addr, *pointer++ );
				EE_Addr += 2;			// address of next word in EEPROM
				_wait_eedata();
		}
	}
} // Write_Alarm

/*
 End of File
*/
